#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Managers for generating flatfiles that combine multiple table into a simplified file.
"""
from __future__ import unicode_literals
from calaccess_processed.postgres import (
    JSONArrayLength,
    JSONExtractPath,
    MaxFromJSONIntegerArray,
)
from django.db.models import F, Q
from django.db.models import Count, Max
from calaccess_processed.managers import BulkLoadSQLManager


class OCDFlatCandidacyManager(BulkLoadSQLManager):
    """
    Custom manager for flattening the contents of the OCD Candidacy model.
    """
    def get_queryset(self):
        """
        Returns the custom QuerySet for this manager.
        """
        return super(
            OCDFlatCandidacyManager, self
        ).get_queryset().filter(
            Q(person__identifiers__scheme='calaccess_filer_id')
            | Q(person__identifiers__isnull=True)
        ).annotate(
            name=F('candidate_name'),
            office=F('post__label'),
            party_name=F('party__name'),
            election_name=F('contest__election__name'),
            election_date=F('contest__election__date'),
            special_election=F('contest__previous_term_unexpired'),
            ocd_person_id=F('person__id'),
            ocd_candidacy_id=F('id'),
            ocd_election_id=F('contest__election'),
            ocd_post_id=F('post__id'),
            ocd_contest_id=F('contest'),
            ocd_party_id=F('party'),
            latest_calaccess_filer_id=Max('person__identifiers__identifier'),
            calaccess_filer_id_count=Count('person__identifiers__identifier'),
            latest_form501_filing_id=MaxFromJSONIntegerArray(
                'extras',
                'form501_filing_ids'
            ),
            form501_filing_count=JSONArrayLength(
                JSONExtractPath('extras', 'form501_filing_ids')
            ),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Models for storing data from Campaign Disclosure Statements (Form 460).
"""
from django.db import models
from calaccess_processed_filings.models.campaign import CampaignExpenditureSubItemBase


class Form460ScheduleGItemBase(CampaignExpenditureSubItemBase):
    """
    Abstract base model for items reported on Schedule G of Form 460 filings.

    On Schedule G, campaign filers are required to itemize payments made on
    their behalf by agents or contractors during the period covered by the
    filing.
    """
    agent_title = models.CharField(
        verbose_name='agent title',
        max_length=10,
        blank=True,
        help_text='Name title of the agent (from EXPN_CD.AGENT_NAMT)',
    )
    agent_lastname = models.CharField(
        verbose_name='agent lastname',
        max_length=200,
        blank=True,
        help_text='Last name of the agent or business name (from '
                  'EXPN_CD.AGENT_NAML)',
    )
    agent_firstname = models.CharField(
        verbose_name='agent firstname',
        max_length=45,
        help_text='First name of the agent (from EXPN_CD.AGENT_NAMF)',
    )
    agent_name_suffix = models.CharField(
        verbose_name='agent name suffix',
        max_length=10,
        blank=True,
        help_text='Name suffix of the agent (from EXPN_CD.AGENT_NAMS)',
    )
    PARENT_SCHEDULE_CHOICES = (
        ('E', 'Schedule E: Payments Made'),
        ('F', 'Schedule F: Accrued Expenses (Unpaid Bills)')
    )
    parent_schedule = models.CharField(
        max_length=1,
        blank=True,
        help_text="Indicates which schedule (E or F) includes the parent item "
                  "(from EXPN_CD.G_FROM_E_F)",
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        abstract = True


class Form460ScheduleGItem(Form460ScheduleGItemBase):
    """
    Payments made by on behalf of campaign filers.

    These transactions are itemized on Schedule G of the most recent version
    of each Form 460 filing. For payments itemized on any version of any Form
    460 filing, see Form460schedulegitemversion.

    Derived from EXPN_CD records where FORM_TYPE is 'G'.
    """
    filing = models.ForeignKey(
        'Form460Filing',
        related_name='schedule_g_items',
        null=True,
        on_delete=models.SET_NULL,
        help_text='Foreign key referring to the Form 460 on which the '
                  'payment was reported (from RCPT_CD.FILING_ID)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing',
            'line_item',
        ),)
        verbose_name = "Form 460 (Campaign Disclosure) Schedule G item"

    def __str__(self):
        return '%s-%s' % (self.filing, self.line_item)


class Form460ScheduleGItemVersion(Form460ScheduleGItemBase):
    """
    Every version of each payment made on behalf of a campaign filer.

    For payments itemized on Schedule G of the most recent version of each Form
    460 filing, see Form460ScheduleGitem.

    Derived from EXPN_CD records where FORM_TYPE is 'G'.
    """
    filing_version = models.ForeignKey(
        'Form460FilingVersion',
        related_name='schedule_g_items',
        null=True,
        on_delete=models.SET_NULL,
        help_text='Foreign key referring to the version of the Form 460 that '
                  'includes the payment made'
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing_version',
            'line_item',
        ),)
        index_together = ((
            'filing_version',
            'line_item',
        ),)
        verbose_name = "Form 460 (Campaign Disclosure) Schedule G item version"

    def __str__(self):
        return '%s-%s-%s' % (
            self.filing_version.filing_id,
            self.filing_version.amend_id,
            self.line_item
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Models for storing data from Campaign Disclosure Statements (Form 461).
"""
from django.db import models
from calaccess_processed_filings.models.campaign import CampaignFinanceFilingBase


class Form461FilingBase(CampaignFinanceFilingBase):
    """
    Base and abstract model for Form 461 filings.
    """
    statement_type = models.CharField(
        max_length=50,
        verbose_name='statement type',
        help_text='Type of statement, e.g., "Quarterly", "Semi-Annual", Pre-'
                  'Election (from CVR_CAMPAIGN_DISCLOSURE.STMT_TYPE)',
    )
    from_date = models.DateField(
        verbose_name='from date',
        db_index=True,
        null=False,
        help_text="The first date of the filing period covered by the statement "
                  "(from CVR_CAMPAIGN_DISCLOSURE.FROM_DATE)",
    )
    thru_date = models.DateField(
        verbose_name='thru date',
        db_index=True,
        null=False,
        help_text="The last date of the filing period covered by the statement "
                  "(from CVR_CAMPAIGN_DISCLOSURE.THRU_DATE)",
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        abstract = True

    def __str__(self):
        return str(self.filing_id)


class Form461Filing(Form461FilingBase):
    """
    The most recent version of each Form 461 filing by independent expenditure or major donor committees.
    """
    filing_id = models.IntegerField(
        verbose_name='filing id',
        primary_key=True,
        null=False,
        help_text='Unique identification number for the Form 461 filing ('
                  'from CVR_CAMPAIGN_DISCLOSURE_CD.FILING_ID)',
    )
    amendment_count = models.IntegerField(
        verbose_name='Count amendments',
        db_index=True,
        null=False,
        help_text='Number of amendments to the Form 461 filing (from '
                  'maximum value of CVR_CAMPAIGN_DISCLOSURE_CD.AMEND_ID)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        index_together = ((
            'filing_id',
            'amendment_count',
        ),)
        verbose_name = "Form 461 (Campaign Disclosure) filing"

    def __str__(self):
        return str(self.filing_id)


class Form461FilingVersion(Form461FilingBase):
    """
    Every version of each Form 461 filing by independent expenditure or major donor committees.
    """
    filing = models.ForeignKey(
        'Form461Filing',
        related_name='versions',
        db_constraint=False,
        null=True,
        on_delete=models.SET_NULL,
        help_text='Unique identification number for the Form 461 filing ('
                  'from CVR_CAMPAIGN_DISCLOSURE_CD.FILING_ID)',
    )
    amend_id = models.IntegerField(
        verbose_name='amendment id',
        null=False,
        help_text='Identifies the version of the Form 461 filing, with 0 '
                  'representing the initial filing (from CVR_CAMPAIGN_'
                  'DISCLOSURE_CD.AMEND_ID)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing',
            'amend_id',
        ),)
        index_together = ((
            'filing',
            'amend_id',
        ),)
        verbose_name = "Form 461 (Campaign Disclosure) filing version"

    def __str__(self):
        return '%s-%s' % (self.filing, self.amend_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Import all of the managers from submodules and thread them together.
"""
from .base import CampaignFinanceManager
from .committees import (
    OCDCommitteeManager,
    OCDCommitteeIdentifierManager,
    OCDCommitteeNameManager,
    OCDCommitteeTypeManager
)
from .filings import (
    OCDFilingManager,
    OCDFilingIdentifierManager,
    OCDFilingActionManager,
    OCDFilingActionSummaryAmountManager
)
from .transactions import OCDTransactionManager


__all__ = (
    "CampaignFinanceManager",
    "OCDCommitteeManager",
    "OCDCommitteeIdentifierManager",
    "OCDCommitteeNameManager",
    "OCDCommitteeTypeManager",
    "OCDFilingManager",
    "OCDFilingIdentifierManager",
    "OCDFilingActionManager",
    "OCDFilingActionSummaryAmountManager",
    "OCDTransactionManager"
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Proxy model for augmenting ScrapedCandidate model with methods useful for processing.
"""
from __future__ import unicode_literals, absolute_import
from django.db.models.functions import Concat
from django.db.models import Value, CharField
from calaccess_processed_elections import corrections

# Models
from .base import ScrapedNameMixin
from .candidateelections import ScrapedCandidateElectionProxy
from calaccess_scraped.models import Candidate
from opencivicdata.elections.models import CandidateContest
from calaccess_processed_elections.proxies import (
    OCDPartyProxy,
    OCDCandidacyProxy,
    OCDPostProxy
)

# Logging
import logging
logger = logging.getLogger(__name__)


class ScrapedCandidateProxy(Candidate, ScrapedNameMixin):
    """
    A proxy for the calaccess_scraped Candidate model.
    """
    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True

    @property
    def election_proxy(self):
        """
        Return the proxy model for the related election object.
        """
        return ScrapedCandidateElectionProxy.objects.get(id=self.election.id)

    @property
    def post_proxy(self):
        """
        Return the proxy model for the related post object.
        """
        post, post_created = OCDPostProxy.objects.get_or_create_by_name(
            self.office_name
        )
        return post

    def get_corrected_party(self):
        """
        Returns a manually correction to the candidate's party, if it's been made. Otherwise returns None.
        """
        scraped_election = self.election_proxy
        return corrections.candidate_party(
            self.name,
            scraped_election.date.year,
            scraped_election.election_type,
            self.office_name,
        )

    def get_party(self):
        """
        Returns the party we believe the candidate was associated with in this election.
        """
        from calaccess_processed_filings.models import Form501Filing

        # First, if the candidate is running for this office, it is by definition non-partisan
        if self.office_name == 'SUPERINTENDENT OF PUBLIC INSTRUCTION':
            logger.debug("{} party set to NO PARTY PREFERENCE based on office".format(self))
            return OCDPartyProxy.objects.get(name="NO PARTY PREFERENCE")

        # Next pull the OCD election record so we have it to inspect
        scraped_election = self.election_proxy

        # Check if this candidate has been manually corrected.
        party = self.get_corrected_party()
        # If so, just pass that out right away
        if party:
            logger.debug("{} party set to {} based on correction".format(self, party))
            return party

        # Next, check if they have filed any Form 501s
        try:
            form501s = self.get_form501s()[0].order_by('-date_filed')
        except Form501Filing.DoesNotExist:
            pass
        else:
            for i in form501s.all():
                # Try getting party from form 501 party
                party = OCDPartyProxy.objects.get_by_name(i.party)
                if not party.is_unknown():
                    logger.debug(
                        "{} party set to {} based on Form 501 party".format(self, party)
                    )
                    return party
                # Try getting it from form 501 filer id
                party = OCDPartyProxy.objects.get_by_filer_id(int(i.filer_id), scraped_election.date)
                if not party.is_unknown():
                    logger.debug("{} party set to {} based on Form 501 filer id".format(self, party))
                    return party

        # If there's no 501, or if the 501 returned an unknown party ...
        # ... try one last stab at using the filer id (assuming it exists)
        if self.scraped_id:
            party = OCDPartyProxy.objects.get_by_filer_id(int(self.scraped_id), scraped_election.date)
            logger.debug("{} party set to {} after checking its scraped filer id".format(self, party))
            return party
        # Otherwise just give up and return the unknown party
        else:
            logger.debug("{} party set to UNKNOWN after failing to find a match".format(self))
            return OCDPartyProxy.objects.get(name="UNKNOWN")

    def match_form501s_by_scraped_id(self):
        """
        Get Form501Filings with filer_ids that match candidate's scraped_id.

        First, filter to matches with the same election_type and office where the
        election_year is the same or earlier than the scraped election's year.

        If none are found, search for matches with the same office office before
        where the election_year is the same or earlier than the election_year.

        Return a Form501Filing QuerySet.
        """
        from calaccess_processed_filings.models import Form501Filing

        election_data = self.election_proxy.parsed_name
        office_data = self.parse_office_name()

        # filter all form501 lookups by office type, district and election year
        # get the most recently filed Form501 within the election_year
        q = Form501Filing.objects.filter(
            office__iexact=office_data['type'],
            district=office_data['district'],
            election_year__lte=election_data['year'],
            filer_id=self.scraped_id,
        )

        # filter to election_type if it will yield results
        if q.filter(election_type=election_data['type']).exists():
            q = q.filter(election_type=election_data['type'])

        return q

    def match_form501s_by_name(self):
        """
        Get Form501Filings with names that match the scraped candidate's name.

        Returns a Form501Filing QuerySet.
        """
        from calaccess_processed_filings.models import Form501Filing

        election_data = self.election_proxy.parsed_name
        office_data = self.parse_office_name()

        q = Form501Filing.objects.filter(
            office__iexact=office_data['type'],
            district=office_data['district'],
            election_year__lte=election_data['year'],
        )
        # first, try "<last_name>, <first_name> <middle_name>" format
        last_first_middle_q = q.annotate(
            full_name=Concat(
                'last_name',
                Value(', '),
                'first_name',
                Value(' '),
                'middle_name',
                output_field=CharField(),
            ),
        ).filter(full_name=self.name)

        # limit scope to election_type if it will yield results
        if last_first_middle_q.filter(election_type=election_data['type']).exists():
            q = last_first_middle_q.filter(election_type=election_data['type'])
        # if not, check if dropping election_type filter yields results
        elif last_first_middle_q.exists():
            q = last_first_middle_q
        # if not, change name format
        else:
            last_first_q = q.annotate(
                full_name=Concat(
                    'last_name',
                    Value(', '),
                    'first_name',
                    output_field=CharField(),
                ),
            ).filter(full_name=self.name)

            # again, limit to election_type at first
            if last_first_q.filter(election_type=election_data['type']).exists():
                q = last_first_q.filter(election_type=election_data['type'])
            else:
                q = last_first_q

        return q

    def get_form501s(self):
        """
        Get the Form501Filings for the scraped candidate.

        If none, raise a Form501Filing.DoesNotExist exception.

        Return a tuple with Form501Filing QuerySet and str describing match criteria.
        """
        from calaccess_processed_filings.models import Form501Filing

        if self.match_form501s_by_scraped_id().exists():
            return (self.match_form501s_by_scraped_id(), 'scraped_id')
        elif self.match_form501s_by_name().exists():
            return (self.match_form501s_by_name(), 'name')
        else:
            raise Form501Filing.DoesNotExist()

    def get_or_create_contest(self):
        """
        Get or create an OCD CandidateContest object.

        Primary contests before 2012 were each linked to a Party. Unless the user
        passes a value via the party keyword argument, the Party linked to the scraped
        candidate's most recent Form501Filing will be selected.

        Returns a tuple (CandidateContest object, created), where created is a boolean
        specifying whether a CandidateContest was created.
        """
        # Get election data
        scraped_election = self.election_proxy

        # Get the candidate's party
        candidate_party = self.get_party()

        # Assume all "SPECIAL" and "RECALL" candidate elections are for contests
        # where the previous term of the office was unexpired.
        if scraped_election.is_special or scraped_election.is_recall:
            previous_term_unexpired = True
            # We are not setting a contest party here for the reasons laid out in the following ticket:
            # https://github.com/california-civic-data-coalition/django-calaccess-processed-data/issues/70#issuecomment-300836502  # NOQA
            contest_party = None
            contest_name = '{} ({})'.format(self.office_name, scraped_election.election_type)
        # Otherwise, we assume this a typical election
        else:
            # At a minimum, that means that the previous term has expired for the office
            previous_term_unexpired = False
            # Beyond that, there are two different cases depending on when the election was held.
            # Prior to 2012, California held partisan primaries. Since then, the state has held
            # open "jungle" primaries that set all candidates for the same office against each other,
            # regardless of their party.
            # In the case of partisan elections, we want to make sure the candidates are separated by party.
            # The one exception to this is superintendent races which have always been non-partisan.
            if scraped_election.is_partisan_primary and self.office_name != 'SUPERINTENDENT OF PUBLIC INSTRUCTION':
                # In this case, the contest party should be the same as the candidate party.
                contest_party = candidate_party
                # If the party is unknown, just tack party on the end of the contest name
                if candidate_party.is_unknown():
                    contest_name = '{} ({} PARTY)'.format(self.office_name, candidate_party.name)
                # For a regular party, we shouldn't have to do much here.
                else:
                    contest_name = '{} ({})'.format(self.office_name, candidate_party.name)
            # If this is a general election prior to 2012, or any non-special election since then ...
            else:
                # ... there is no party for the contest ...
                contest_party = None
                # ... and there's no need to do anything to the contest name.
                contest_name = self.office_name

        # Make it happen
        contest, created = CandidateContest.objects.get_or_create(
            election=scraped_election.get_ocd_election(),
            name=contest_name,
            previous_term_unexpired=previous_term_unexpired,
            party=contest_party,
            division=self.post_proxy.division,
        )

        # if contest was created, add the Post
        if created:
            contest.posts.create(post=self.post_proxy)

        # Always update the source for the contest
        contest.sources.update_or_create(
            url=self.url,
            note='Last scraped on {dt:%Y-%m-%d}'.format(dt=self.last_modified)
        )

        # Return the contest and whether or not it was created
        return contest, created

    def get_loaded_ocd_candidacy(self):
        """
        Get an OCD candidacy previously loaded from the scraped Candidate.

        Matches are scoped to OCD Candidacies where:
        - The OCD Election linked to the scraped election's id
        - The Post that's equivalent to the scrape candidate's office_name
        - If the scraped election is "special", OCD CandidateContest where
        the previous term is unexpired.
        - If the scraped election is "regular", OCD CandidateContest where
        the previous term is expired.

        If the scraped candidate has a filer_id, use this in the matching
        criteria. Otherwise, match on candidate_name, person.name or
        person__other_name__name.

        Returns a Candidacy object.
        """
        scraped_election = self.election_proxy
        ocd_election = scraped_election.get_ocd_election()

        q = OCDCandidacyProxy.objects.filter(
            contest__election=ocd_election,
            post=self.post_proxy,
        )

        # for special elections, filter to contests with unexpired terms
        if scraped_election.is_special or scraped_election.is_recall:
            q = q.filter(contest__previous_term_unexpired=True)
        # for regular elections, filter to contests with expired_terms
        elif scraped_election.is_regular:
            q = q.filter(contest__previous_term_unexpired=False)
        else:
            raise Exception(
                'Unknown election type for %s.' % scraped_election
            )

        if self.scraped_id:
            candidacy = q.get_by_filer_id(self.scraped_id)
        else:
            candidacy = q.get_by_name(self.parsed_name['name'])

        return candidacy



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Proxy models for augmenting our source data tables with methods useful for processing.
"""
from .divisions import OCDDivisionProxy
from .organizations import (
    OCDMembershipProxy,
    OCDOrganizationProxy,
    OCDOrganizationIdentifierProxy,
    OCDOrganizationNameProxy,
)
from .jurisdictions import OCDJurisdictionProxy
from .people import (
    OCDPersonProxy,
    OCDPersonIdentifierProxy,
    OCDPersonNameProxy,
)
from .posts import OCDPostProxy


__all__ = (
    "OCDDivisionProxy",
    "OCDMembershipProxy",
    "OCDOrganizationProxy",
    "OCDOrganizationIdentifierProxy",
    "OCDOrganizationNameProxy",
    "OCDJurisdictionProxy"
    "OCDPersonProxy",
    "OCDPersonIdentifierProxy",
    "OCDPersonNameProxy",
    "OCDPostProxy",
    "OCDJurisdictionProxy",
    "OCDPersonProxy"
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Load the OCD Candidacy model with data extracted from the Form501Filing model.
"""
from opencivicdata.elections.models import CandidateContest
from calaccess_processed_filings.models import Form501Filing
from calaccess_processed_elections.proxies import OCDCandidacyProxy
from calaccess_processed.management.commands import CalAccessCommand


class Command(CalAccessCommand):
    """
    Load the OCD Candidacy model with data extracted from the Form501Filing model.
    """
    help = 'Load the OCD Candidacy model with data extracted from the Form501Filing model'

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(Command, self).handle(*args, **options)

        if not CandidateContest.objects.exists():
            error_message = 'No contests currently loaded (run loadocdcandidatecontests).'
            self.log(error_message)
        else:
            form501_count = Form501Filing.objects.without_candidacy().count()
            self.header(f"Processing {form501_count} Form 501 filings without candidacies")
            self.load()
            self.success("Done!")

    def load(self):
        """
        Loop over each Form501Filing and attempt to load into OCD models.
        """
        for form501 in Form501Filing.objects.without_candidacy():
            if self.verbosity > 2:
                self.log(' Processing Form 501: %s' % form501.filing_id)
                self.process_form501(form501)

    def process_form501(self, form501):
        """
        Process a Form501Filing and attempt to load its data into OCD models.
        """
        # Get a linked contest
        contest = form501.get_or_create_contest()

        # If there is no contest, skip.
        if not contest:
            pass
        else:
            candidacy, created = OCDCandidacyProxy.objects.get_or_create_from_calaccess(
                contest,
                form501.parsed_name,
                candidate_filer_id=form501.filer_id
            )

            if created and self.verbosity > 2:
                self.log(' Created Candidacy: %s' % candidacy)

            candidacy.link_form501(form501.filing_id)
            candidacy.update_from_form501()
            candidacy.update_party_from_form501()

        return form501



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Proxy models for augmenting our source data tables with methods useful for processing.
"""
from __future__ import unicode_literals
from django.db.models.functions import Cast

# Models
from django.db.models import (
    IntegerField,
    Case,
    Q,
    When
)
from .elections import OCDElectionProxy
from opencivicdata.core.models import Membership
from opencivicdata.elections.models import Candidacy, CandidacySource
from calaccess_processed.proxies import OCDProxyModelMixin
from ..core.people import OCDPersonProxy

# Managers
from calaccess_processed.managers import BulkLoadSQLManager
from calaccess_processed_elections.managers import OCDCandidacyManager


class OCDCandidacyProxy(Candidacy, OCDProxyModelMixin):
    """
    A proxy on the OCD Candidacy model with helper methods.
    """
    objects = OCDCandidacyManager()

    copy_to_fields = (
        ('id',),
        ('candidate_name',),
        ('person',),
        ('party',),
        ('contest',),
        ('post',),
        ('is_incumbent',),
        ('registration_status',),
        ('top_ticket_candidacy',),
        ('filed_date',),
        ('created_at',),
        ('updated_at',),
        ('extras',),
        ('locked_fields',),
    )

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True

    @property
    def election_proxy(self):
        """
        Returns the proxied OCDElectionProxy linked to this candidacy.
        """
        return OCDElectionProxy.objects.get(id=self.contest.election_id)

    @property
    def filer_ids(self):
        """
        Returns the CAL-ACCESS filer_id linked with the object, if any.
        """
        return self.person.identifiers.filter(scheme="calaccess_filer_id")

    @property
    def form501_filing_ids(self):
        """
        Returns any linked Form 501 filing ids.
        """
        try:
            return self.extras['form501_filing_ids']
        except KeyError:
            return []

    @property
    def form501s(self):
        """
        Returns any linked Form 501 objects.
        """
        from calaccess_processed_filings.models import Form501Filing
        return Form501Filing.objects.filter(filing_id__in=self.form501_filing_ids)

    @property
    def person_proxy(self):
        """
        Returns an OCDPersonProxy instance linked to the Candidacy.
        """
        person = self.person
        person.__class__ = OCDPersonProxy
        return person

    def link_form501(self, form501_id):
        """
        Link an id of a Form501Filing to a Candidacy, if it isn't already.
        """
        # Check if the attribute is already there
        if 'form501_filing_ids' in self.extras:
            # If it is, check if we already have this id
            if form501_id not in self.extras['form501_filing_ids']:
                # If we don't, append it to the list
                self.extras['form501_filing_ids'].append(form501_id)
                # Save out
                self.save()
        # If the attribute isn't there, go ahead and add it.
        else:
            self.extras['form501_filing_ids'] = [form501_id]
            # Save out
            self.save()

    def update_from_form501(self):
        """
        Set Candidacy fields using data extracted from linked Form501Filings.
        """
        from calaccess_processed_filings.models import Form501Filing

        # get all Form501Filing linked to Candidacy
        filing_ids = self.extras['form501_filing_ids']
        filings = Form501Filing.objects.filter(filing_id__in=filing_ids)

        # keep the earliest filed_date
        first_filed_date = filings.earliest('date_filed').date_filed

        # Update filed_date if not the earliest
        if self.filed_date != first_filed_date:
            self.filed_date = first_filed_date
            self.save()

        # set registration status to "withdrawn" based on statement_type of latest Form501
        latest = filings.latest('date_filed')
        if latest.statement_type == '10003':  # <-- This is the code for withdrawn
            # If the candidacy hasn't been marked that way, update it now
            if self.registration_status != 'withdrawn':
                self.registration_status = 'withdrawn'
                self.save()

    def link_filer_ids_from_form501s(self):
        """
        Create PersonIdentifiers for each filer_id from Form501Filings.
        """
        from calaccess_processed_filings.models import Form501Filing
        person = self.person
        current_filer_ids = [
            i.identifier for i in person.identifiers.filter(scheme='calaccess_filer_id')
        ]

        filing_ids = self.extras['form501_filing_ids']
        missing_filer_ids = [
            f.filer_id for f in Form501Filing.objects.filter(
                filing_id__in=filing_ids
            ).exclude(
                filer_id__in=current_filer_ids
            )
        ]
        for i in missing_filer_ids:
            person.identifiers.get_or_create(
                scheme='calaccess_filer_id',
                identifier=i,
            )

    def update_party_from_form501(self):
        """
        Update party for Candidacy based on latest Form501 where its populated.
        """
        from calaccess_processed_filings.models import Form501Filing

        # get all Form501Filing linked to Candidacy
        filing_ids = self.extras['form501_filing_ids']
        filings = Form501Filing.objects.filter(filing_id__in=filing_ids)

        latest_party = filings.filter(
            party__isnull=False
        ).latest('date_filed').get_party()

        if latest_party != self.party:
            self.party = latest_party
            self.save()

    def check_incumbency(self):
        """
        Check if the Candidacy is for the incumbent officeholder.

        Return True if:
        * Membership exists for the Person and Post linked to the Candidacy, and
        * Membership.end_date is NULL or has a year later than Election.date.year.
        """
        incumbent_q = Membership.objects.filter(
            post=self.post,
            person=self.person,
        ).annotate(
            # Cast end_date's value as an int, treat '' as NULL
            end_year=Cast(
                Case(When(end_date='', then=None)),
                IntegerField(),
            )
        ).filter(
            Q(end_year__gt=self.election.date.year)
            | Q(end_date='')
        )
        if incumbent_q.exists():
            return True
        else:
            return False


class OCDCandidacySourceProxy(CandidacySource, OCDProxyModelMixin):
    """
    A proxy on the OCD CandidacySource model.
    """
    objects = BulkLoadSQLManager()

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Models for storing campaign-related entities derived from raw CAL-ACCESS data.
"""
from datetime import date
from calaccess_processed_elections import corrections, get_expected_election_date

# Managers
from calaccess_processed_filings.managers import Form501FilingManager

# Models
from django.db import models
from opencivicdata.elections.models import CandidateContest
from calaccess_processed_filings.models.base import FilingBaseModel


class Form501FilingBase(FilingBaseModel):
    """
    Base and abstract model for Form 501 filings.
    """
    date_filed = models.DateField(
        verbose_name='from date',
        null=True,
        db_index=True,
        help_text="Date when the Form 501 filing was filed (from F501_502_CD.RPT_DATE)",
    )
    statement_type = models.CharField(
        max_length=62,
        verbose_name='statement type',
        help_text='Describes the type of statement, e.g. "ORIGINAL", "AMENDMENT" '
                  '(from LOOKUP_CODES.CODE_DESC)',
    )
    filer_id = models.CharField(
        verbose_name="filer identifier",
        max_length=9,
        help_text="Filer's unique identifier (from F501_502_CD.FILER_ID)",
    )
    committee_id = models.CharField(
        verbose_name='committee identifier',
        max_length=9,
        help_text="Candidate's committee's unique filer idenitifier (from "
                  "F501_502_CD.COMMITTEE_ID)",
    )
    title = models.CharField(
        verbose_name="candidate name title",
        max_length=100,
        blank=True,
        help_text="Name title of the candidate (from F501_502_CD.CAND_NAMT)",
    )
    last_name = models.CharField(
        verbose_name="candidate last name",
        max_length=200,
        # just a few don't even have a last name
        blank=True,
        help_text="Last name of the candidate (from F501_502_CD.CAND_NAML)",
    )
    first_name = models.CharField(
        verbose_name="candidate first name",
        max_length=45,
        blank=True,
        help_text="First name of the candidate (from F501_502_CD.CAND_NAMF)",
    )
    middle_name = models.CharField(
        verbose_name="candidate middle name",
        max_length=20,
        blank=True,
        help_text="Middle name of the candidate (from F501_502_CD.CAND_NAMM)",
    )
    name_suffix = models.CharField(
        verbose_name="candidate name suffix",
        max_length=10,
        blank=True,
        help_text="Name suffix of the candidate (from F501_502_CD.CAND_NAMS)",
    )
    name_moniker = models.CharField(
        verbose_name="candidate name moniker",
        max_length=20,
        blank=True,
        help_text="Moniker (aka, nickname) of the candidate (from F501_502_CD"
                  ".MONIKER)",
    )
    phone = models.CharField(
        max_length=20,
        verbose_name='candidate phone number',
        blank=True,
        help_text="Phone number of the candidate (from F501_502_CD.CAND_PHON)",
    )
    fax = models.CharField(
        max_length=20,
        verbose_name='fax number',
        blank=True,
        help_text="Phone number of the candidate (from F501_502_CD.CAND_FAX)",
    )
    email = models.CharField(
        max_length=200,
        verbose_name='email address',
        blank=True,
        help_text="Email address of the candidate (from F501_502_CD.CAND_EMAIL)",
    )
    city = models.CharField(
        max_length=200,
        verbose_name="candidate city",
        blank=True,
        help_text="City of the candidate (from F501_502_CD.CAND_CITY)",
    )
    state = models.CharField(
        max_length=200,
        verbose_name='candidate state',
        blank=True,
        help_text="State of the candidate (from F501_502_CD.CAND_ST)",
    )
    zip_code = models.CharField(
        max_length=10,
        verbose_name='zip code',
        blank=True,
        help_text='Zip code (usually zip5, sometimes zip9) of the '
                  'candidate (from F501_502_CD.CAND_ZIP4)',
    )
    office = models.CharField(
        verbose_name='office sought',
        max_length=80,
        blank=True,
        help_text='Position title of the office sought by the candidate (from '
                  'LOOKUP_CODES_CD.CODE_DESC, unless NULL or 0, then F501_502_CD.'
                  'OFFICE_DSCR)',
    )
    agency = models.CharField(
        verbose_name='agency name',
        max_length=200,
        blank=True,
        help_text='Name of the agency with the office sought (from '
                  'F501_502_CD.AGENCY_NAM)',
    )
    district = models.IntegerField(
        verbose_name='district',
        null=True,
        help_text='District of office sought, if applicable (from LOOKUP_CODES_CD'
                  '.CODE_DESC, unless NULL or 0, then F501_502_CD.DIST_NO)',
    )
    party = models.CharField(
        max_length=30,
        verbose_name='political party',
        blank=True,
        help_text='Political party of the candidate (from LOOKUP_CODES_CD.'
                  'CODE_DESC, unless NULL or 0, then F501_502_CD.PARTY)',
    )
    jurisdiction = models.CharField(
        max_length=30,
        verbose_name='jurisdiction',
        blank=True,
        help_text='Jurisdiction of the office sought, e.g., "LOCAL", "STATE" '
                  '(from LOOKUP_CODES_CD.CODE_DESC)',
    )
    election_type = models.CharField(
        verbose_name='election type',
        max_length=16,
        null=True,
        help_text='Type of election in which the candidate is declaring intention'
                  ' to run, e.g. "PRIMARY", "GENERAL" (from LOOKUP_CODES_CD.'
                  'CODE_DESC)',
    )
    election_year = models.IntegerField(
        verbose_name='election year',
        null=True,
        help_text='Year in which the election is held (from F501_502_CD.YR_OF_ELEC)',
    )
    accepted_limit = models.BooleanField(
        null=True,
        help_text='Indicates if either the "I accept the voluntary expenditure '
                  'ceiling" or "I do not accept the voluntary expenditure" '
                  'box is checked (from F501_502_CD.ACCEPT_LIMIT_YN)',
    )
    limit_not_exceeded_election_date = models.DateField(
        verbose_name='limit not exceeded election date',
        null=True,
        help_text='Date of the primary or special election in which the candidate '
                  'did not accept the voluntary expenditure ceiling but also did '
                  'exceed the ceiling. Candidates may amend their Form 501 to accept '
                  'the limits for the general election or special election runoff '
                  'and receive all the benefits of accepting the ceiling (from '
                  'F501_502_CD.DID_EXCEED_DT)'
    )
    personal_funds_contrib_date = models.DateField(
        verbose_name='personal funds contribution date',
        null=True,
        help_text='Date on which the candidate contributed personal funds in excess '
                  'of the voluntary expenditure ceiling for the (from F501_502_CD'
                  '.CNTRB_PRSNL_FNDS_DT)',
    )
    executed_on = models.DateField(
        verbose_name='executed on date',
        null=True,
        help_text='Date on which the candidate intention statement was signed '
                  '(from F501_502_CD.EXECUTE_DT)'
    )

    class Meta:
        """
        Model options.
        """
        abstract = True
        app_label = 'calaccess_processed_filings'


class Form501Filing(Form501FilingBase):
    """
    The most recent version of each Form 501 filing by a candidate.

    Includes information from the most recent version of each Form 501 filing.
    All versions of the filings can be found in Form501FilingVersion.
    """
    filing_id = models.IntegerField(
        verbose_name='filing id',
        primary_key=True,
        null=False,
        help_text='Unique identification number for the Form 501 filing ('
                  'from F501_502_CD.FILING_ID)',
    )
    amendment_count = models.IntegerField(
        verbose_name='Count amendments',
        db_index=True,
        null=False,
        help_text='Number of amendments to the Form 501 filing (from '
                  'maximum value of F501_502_CD.AMEND_ID)',
    )
    objects = Form501FilingManager()

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        index_together = ((
            'filing_id',
            'amendment_count',
        ),)
        verbose_name = "Form 501 (Candidate Intention) filing"

    def __str__(self):
        return str(self.filing_id)

    @property
    def name(self):
        """
        Return the 'name' of the candidate to match the format we typically put in the OCD Person model.
        """
        split_name = self.sort_name.split(',')
        split_name.reverse()
        return ' '.join(split_name).strip()

    @property
    def sort_name(self):
        """
        Return the 'sort_name' of the candidate to match the format we typically scrape from the CAL-ACCESS website.

        This is useful when trying to consolidate these forms with scraped data in our OCD models.
        """
        return '{0.last_name}, {0.first_name} {0.middle_name}'.format(self).strip()

    @property
    def parsed_name(self):
        """
        The parsed name of the candidate ready to be converted into an OCD Person.
        """
        return dict(
            name=self.name,
            sort_name=self.name
        )

    @property
    def office_name(self):
        """
        Return the 'office_name' of the candidate to match the format we typically scrape from the CAL-ACCESS website.

        This is useful when trying to consolidate these forms with scraped data in our OCD models.
        """
        if self.district:
            return '{0.office} {0.district}'.format(self).strip()
        else:
            return self.office

    @property
    def ocd_election(self):
        """
        Return Election occurring in year with name in including election_type.

        Return None if none found.
        """
        from calaccess_processed_elections.proxies import OCDElectionProxy

        if not self.election_year or not self.election_type:
            return None
        try:
            return OCDElectionProxy.objects.get(
                date__year=self.election_year,
                name__contains=self.election_type,
            )
        except (OCDElectionProxy.DoesNotExist, OCDElectionProxy.MultipleObjectsReturned):
            # if it's a future primary, try to calculate the date
            if self.election_year >= date.today().year and self.election_type == 'PRIMARY':
                try:
                    dt_obj = get_expected_election_date(
                        self.election_year, self.election_type
                    )
                except ValueError:
                    return None
                return OCDElectionProxy.objects.create_from_calaccess(
                    '{0} {1}'.format(self.election_year, self.election_type),
                    dt_obj,
                    election_type=self.election_type,
                )
            else:
                return None

    def get_party(self):
        """
        Get the Party from Form501Filing.

        Return Party object or None.
        """
        from calaccess_processed_elections.proxies import OCDPartyProxy

        # first try the corrections
        party = corrections.candidate_party(
            '{0.last_name}, {0.first_name} {0.middle_name}'.format(self).strip(),
            self.election_year,
            self.election_type,
            '{0.office} {0.district}'.format(self).strip().upper(),
        )
        if party:
            return party

        # then try using the party on the form501
        party = OCDPartyProxy.objects.get_by_name(self.party)
        if not party.is_unknown():
            return party

        # finally, try looking in FilerToFilerTypes
        ocd_election = self.ocd_election
        if not ocd_election:
            return OCDPartyProxy.objects.unknown()
        return OCDPartyProxy.objects.get_by_filer_id(int(self.filer_id), ocd_election.date)

    def get_or_create_contest(self):
        """
        Get or create a CandidateContest by extracting info form Form501Filing.

        Return a CandidateContest or None, if extracted info is insufficient.
        """
        from calaccess_processed_elections.proxies import OCDPostProxy

        # Get or create an election
        ocd_election = self.ocd_election
        if not ocd_election:
            return None

        # Get or create a post
        post = OCDPostProxy.objects.get_by_form501(self)

        # Don't bother trying to get contest unless we have a post
        if not post:
            return None

        # Seed contest data
        contest_data = {
            'posts__post': post,
            'division': post.division,
            'election': ocd_election,
        }

        # if looking for a pre-2012 primary, include party
        if ocd_election.is_partisan_primary:
            contest_data['party'] = self.get_party()

        # Try to get it from the database
        try:
            return CandidateContest.objects.get(**contest_data)
        except CandidateContest.DoesNotExist:
            # if the election date is later than today, but no contest
            if ocd_election.date > date.today():
                # make the contest (CAL-ACCESS website might behind)
                contest = CandidateContest.objects.create(
                    name=post.label.upper(),
                    division=contest_data['division'],
                    election=contest_data['election'],
                )
                contest.posts.create(
                    contest=contest,
                    post=contest_data['posts__post'],
                )
                return contest
            # Otherwise give up
            else:
                return None
        except CandidateContest.MultipleObjectsReturned:
            # In this case, there is likely a primary and a runoff on the same day
            # and the code is unable to specify between them.
            # I don't yet have a solution to this problem so we are going to give up
            return None


class Form501FilingVersion(Form501FilingBase):
    """
    Every version of each Form 501 (Candidate Intention Statement) filing by candidates.

    Includes information found on each version of each Form 501 filing. For the
    most recent version of each filing, see Form501Filing.
    """
    filing = models.ForeignKey(
        'Form501Filing',
        related_name='versions',
        db_constraint=False,
        null=True,
        on_delete=models.SET_NULL,
        help_text='Unique identification number for the Form 501 filing ('
                  'from F501_502_CD.FILING_ID)',
    )
    amend_id = models.IntegerField(
        verbose_name='amendment id',
        null=False,
        help_text='Identifies the version of the Form 501 filing, with 0 '
                  'representing the initial filing (from F501_502_CD.FILING_ID)',
    )
    objects = Form501FilingManager()

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing',
            'amend_id',
        ),)
        index_together = ((
            'filing',
            'amend_id',
        ),)
        verbose_name = "Form 501 (Candidate Intention) filing version"

    def __str__(self):
        return '{}-{}'.format(self.filing, self.amend_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Find and merge OCD Person records that share the same CAL-ACCESS filer_id.
"""
from django.db.models import Count
from calaccess_processed_elections.merge import merge_persons
from opencivicdata.core.models import Person, PersonIdentifier
from calaccess_processed.management.commands import CalAccessCommand


class Command(CalAccessCommand):
    """
    Find and merge OCD Person records that share the same CAL-ACCESS filer_id.
    """
    help = 'Find and merge OCD Person records that share the same CAL-ACCESS filer_id'

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(Command, self).handle(*args, **options)

        # Loop over all CAL-ACCESS filer_ids linked to multiple Persons
        shared_filer_ids_q = PersonIdentifier.objects.values(
            'identifier',
        ).annotate(
            person_count=Count('person'),
        ).filter(
            scheme='calaccess_filer_id',
            person_count__gt=1,
        ).order_by()

        if shared_filer_ids_q.count() == 0:
            self.log("No persons to merge by CAL-ACCESS filer_id")
        else:

            self.header(
                "Merging %s Person sets with shared CAL-ACCESS filer_id" % shared_filer_ids_q.count()
            )

            for group in shared_filer_ids_q.all():
                persons = [
                    p for p in Person.objects.filter(
                        identifiers__scheme='calaccess_filer_id',
                        identifiers__identifier=group['identifier'],
                    ).all()
                ]

                if self.verbosity > 2:
                    self.log(
                        "Merging {0} Persons sharing filer_id {1}".format(
                            len(persons),
                            group['identifier'],
                        )
                    )

                merge_persons(persons)

            self.success("Done!")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Models for storing Part 2 data from Schedule 496, the Late Independent Expenditure Reports.

More about the filing: https://calaccess.californiacivicdata.org/documentation/calaccess-forms/f496/
"""
from django.db import models
from calaccess_processed_filings.models.base import FilingBaseModel


class Form496Part2ItemBase(FilingBaseModel):
    """
    Abstract base model for items reported on Part 2 of Schedule 496 filings.

    On Part 2 of Schedule 496, campaign filers are required to report
    independent expenditures made in the 90 days leading up to an election.
    """
    line_item = models.IntegerField(
        verbose_name='line item',
        db_index=True,
        null=False,
        help_text='Line number of the filing form where the transaction is '
                  'itemized (from S496_CD.LINE_ITEM)',
    )
    expense_date = models.DateField(
        verbose_name="expense date",
        null=True,
        help_text="Date or expense (from S496_CD.EXP_DATE)",
    )
    amount = models.DecimalField(
        verbose_name='amount',
        decimal_places=2,
        max_digits=14,
        help_text="Amount paid to the payee in the period covered by the "
                  "filing (from S496_CD.AMOUNT)",
    )
    transaction_id = models.CharField(
        verbose_name='transaction id',
        max_length=20,
        db_index=True,
        help_text='Identifies a unique transaction across versions of the a '
                  'given Schedule 496 filing (from S496_CD.TRAN_ID)'
    )
    payment_description = models.CharField(
        verbose_name="payment description",
        max_length=400,
        blank=True,
        help_text="Purpose of payment and/or description/explanation (from S496_CD.EXPN_DSCR)",
    )
    memo_code = models.CharField(
        verbose_name='memo code',
        max_length=500,
        blank=True,
        help_text="A description offered by the filer",
    )
    memo_reference_number = models.CharField(
        verbose_name='memo reference number',
        max_length=20,
        blank=True,
        help_text='Reference number for the memo attached to the transaction '
                  '(from S496_CD.MEMO_REFNO)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        abstract = True


class Form496Part2Item(Form496Part2ItemBase):
    """
    Late independent expenditures made by campaign filers.

    These transactions are itemized on Part 2 of the most recent version
    of each Schedule 496 filing.
    """
    filing = models.ForeignKey(
        'Form496Filing',
        related_name='independent_expenditures_made',
        db_constraint=False,
        null=True,
        on_delete=models.SET_NULL,
        help_text='Unique identification number for the Schedule 496 filing (from S496_CD.FILING_ID)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing',
            'line_item',
        ),)
        verbose_name = "Form 496 (Late Independent Expenditure) Part 2 item"

    def __str__(self):
        return '%s-%s' % (self.filing, self.line_item)


class Form496Part2ItemVersion(Form496Part2ItemBase):
    """
    Every version of each late independent expenditure made by a campaign filer.
    """
    filing_version = models.ForeignKey(
        'Form496FilingVersion',
        related_name='independent_expenditures_made',
        null=True,
        on_delete=models.SET_NULL,
        help_text='Foreign key referring to the version of the Schedule 496 that includes the given expenditure'
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing_version',
            'line_item',
        ),)
        index_together = ((
            'filing_version',
            'line_item',
        ),)
        verbose_name = "Form 496 (Late Independent Expenditure) Part 2 item version"

    def __str__(self):
        return '%s-%s-%s' % (
            self.filing_version.filing_id,
            self.filing_version.amend_id,
            self.line_item
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Models for storing data from Campaign Disclosure Statements (Form 460).
"""
from django.db import models
from calaccess_processed_filings.models.base import FilingBaseModel


class Form460ScheduleFItemBase(FilingBaseModel):
    """
    Abstract base model for items reported on Schedule F of Form 460 filings.

    On Schedule F, campaign filers report unpaid bills for goods or services
    accrued during the period covered by the filing.
    """
    line_item = models.IntegerField(
        verbose_name='line item',
        help_text="Line number of the filing form where the unpaid bill is "
                  "itemized (from DEBT_CD.LINE_ITEM)",
    )
    PAYEE_CODE_CHOICES = (
        ('BNM', "Ballot measure's name/title"),
        ('COM', "Committee"),
        ('IND', "Individual"),
        ('OTH', "Other"),
        ('PTY', "Political Party"),
        ('RCP', "Recipient committee"),
        ('SCC', "Small Contributor Committee"),
        ('???', "Unknown value"),
    )
    payee_code = models.CharField(
        verbose_name='payee code',
        max_length=3,
        blank=True,
        choices=PAYEE_CODE_CHOICES,
        help_text='Code describing the payee (from DEBT_CD.ENTITY_CD)',
    )
    payee_committee_id = models.CharField(
        verbose_name='committee id',
        max_length=9,
        blank=True,
        help_text="Payee's filer identification number, if it is a "
                  "committee (from DEBT_CD.CMTE_ID)",
    )
    payee_title = models.CharField(
        verbose_name='payee title',
        max_length=10,
        blank=True,
        help_text='Name title of the payee (from DEBT_CD.PAYEE_NAMT)',
    )
    payee_lastname = models.CharField(
        verbose_name='payee lastname',
        max_length=200,
        blank=True,
        help_text='Last name of the payee or business name (from '
                  'DEBT_CD.PAYEE_NAML)',
    )
    payee_firstname = models.CharField(
        verbose_name='payee firstname',
        max_length=45,
        help_text='First name of the payee (from DEBT_CD.PAYEE_NAMF)',
    )
    payee_name_suffix = models.CharField(
        verbose_name='payee name suffix',
        max_length=10,
        blank=True,
        help_text='Name suffix of the payee (from DEBT_CD.PAYEE_NAMS)',
    )
    payee_city = models.CharField(
        verbose_name='payee city',
        max_length=30,
        blank=True,
        help_text='City of the payee (from DEBT_CD.PAYEE_CITY)',
    )
    payee_state = models.CharField(
        verbose_name='payee state',
        max_length=2,
        blank=True,
        help_text='State of the payee (from DEBT_CD.PAYEE_ST)',
    )
    payee_zip = models.CharField(
        verbose_name='payee zip',
        max_length=10,
        blank=True,
        help_text='Zip code (usually zip5, sometimes zip9) of the '
                  'payee (from DEBT_CD.PAYEE_ZIP4)',
    )
    treasurer_title = models.CharField(
        verbose_name='treasurer title',
        max_length=10,
        blank=True,
        help_text="Name title of the payee committee's treasurer (from "
                  "DEBT_CD.TRES_NAMT)",
    )
    treasurer_lastname = models.CharField(
        verbose_name='treasurer lastname',
        max_length=200,
        blank=True,
        help_text="Last name of the payee committee's treasurer (from "
                  "DEBT_CD.TRES_NAML)",
    )
    treasurer_firstname = models.CharField(
        verbose_name='treasurer firstname',
        max_length=45,
        help_text="First name of the payee committee's treasurer (from "
                  "DEBT_CD.TRES_NAMF)",
    )
    treasurer_name_suffix = models.CharField(
        verbose_name='treasurer name suffix',
        max_length=10,
        blank=True,
        help_text="Name suffix of the payee committee's treasurer (from "
                  "DEBT_CD.TRES_NAMS)",
    )
    treasurer_city = models.CharField(
        verbose_name='treasurer city',
        max_length=30,
        blank=True,
        help_text="City of the payee committee's treasurer (from DEBT_CD."
                  "TRES_CITY)",
    )
    treasurer_state = models.CharField(
        verbose_name='treasurer state',
        max_length=2,
        blank=True,
        help_text="State of the payee committee's treasurer (from DEBT_CD."
                  "TRES_ST)",
    )
    PAYMENT_CODE_CHOICES = (
        ('CMP', 'Campaign paraphernalia/miscellaneous'),
        ('CNS', 'Campaign consultants'),
        ('CTB', 'Contribution (if nonmonetary, explain)*'),
        ('CVC', 'Civic donations'),
        ('FIL', 'Candidate filing/ballot feeds'),
        ('FND', 'Fundraising events'),
        ('IKD', 'In-kind contribution (nonmonetary)'),
        ('IND', 'Independent expenditure supporting/opposing others (explain)*'),
        ('LEG', 'Legal defense'),
        ('LIT', 'Campaign literature and mailings'),
        ('LON', 'Loan'),
        ('MBR', 'Member communications'),
        ('MON', 'Monetary contribution'),
        ('MTG', 'Meetings and appearances'),
        ('OFC', 'Office expenses'),
        ('PET', 'Petition circulating'),
        ('PHO', 'Phone banks'),
        ('POL', 'Polling and survey research'),
        ('POS', 'Postage, delivery and messenger services'),
        ('PRO', 'Professional services (legal, accounting)'),
        ('PRT', 'Print ads'),
        ('RAD', 'Radio airtime and production costs'),
        ('RFD', 'Returned contributions'),
        ('SAL', 'Campaign workers salaries'),
        ('TEL', 'T.V. or cable airtime and production costs'),
        ('TRC', 'Candidate travel, lodging and meals (explain)'),
        ('TRS', 'Staff/spouse travel, lodging and meals (explain)'),
        ('TSF', 'Transfer between committees of the same candidate/sponsor'),
        ('VOT', 'Voter registration'),
        ('WEB', 'Information technology costs (internet, e-mail)'),
        ('???', "Unknown value"),
    )
    payment_code = models.CharField(
        verbose_name='payment code',
        max_length=3,
        blank=True,
        choices=PAYMENT_CODE_CHOICES,
        help_text='Code describing the payment (from DEBT_CD.EXPN_CODE)',
    )
    payment_description = models.CharField(
        verbose_name="payment description",
        max_length=400,
        blank=True,
        help_text="Purpose of payment and/or description/explanation (from "
                  "DEBT_CD.EXPN_DSCR)",
    )
    begin_balance = models.DecimalField(
        verbose_name="begin balance",
        decimal_places=2,
        max_digits=14,
        help_text="Outstanding balance at the beginning of period covered by "
                  "the filing (from DEBT_CD.BEG_BAL)",
    )
    amount_paid = models.DecimalField(
        decimal_places=2,
        max_digits=14,
        help_text='Amount paid this period (from DEBT_CD.AMT_PAID)'
    )
    amount_incurred = models.DecimalField(
        verbose_name="amount incurred",
        decimal_places=2,
        max_digits=14,
        help_text='Amount incurred this period (from DEBT_CD.AMT_INCUR)',
    )
    end_balance = models.DecimalField(
        verbose_name="end balance",
        decimal_places=2,
        max_digits=14,
        help_text="Outstanding balance at the end of period covered by the "
                  "filing (from DEBT_CD.END_BAL)",
    )
    transaction_id = models.CharField(
        verbose_name='transaction id',
        max_length=20,
        help_text='Identifies a unique transaction across versions of the a '
                  'given Form 460 filing (from DEBT_CD.TRAN_ID)',
    )
    parent_transaction_id = models.CharField(
        verbose_name='parent transaction id',
        max_length=20,
        blank=True,
        help_text='Refers to a parent transaction itemized on the same Form '
                  '460 filing (from DEBT_CD.BAKREF_TID)',
    )
    memo_reference_number = models.CharField(
        verbose_name='memo reference number',
        max_length=20,
        blank=True,
        help_text="A value assigned by the filer which refers to the item's"
                  "footnote in the TEXT_MEMO_CD table (from DEBT_CD."
                  "MEMO_REFNO)",
    )
    memo_code = models.BooleanField(
        verbose_name='memo_code',
        default=False,
        help_text="Memo amount flag (from DEBT_CD.MEMO_CODE)"
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        abstract = True


class Form460ScheduleFItem(Form460ScheduleFItemBase):
    """
    Accrued expenses of campaign filers.

    These transactions are itemized on Schedule F of the most recent version
    of each Form 460 filing. For accrued expenses itemized on any version of
    of any Form 460 filing, see Form460ScheduleFItemVersion.

    Derived from DEBT_CD records.
    """
    filing = models.ForeignKey(
        'Form460Filing',
        related_name='schedule_f_items',
        null=True,
        on_delete=models.SET_NULL,
        help_text='Foreign key referring to the Form 460 on which the '
                  'payment was reported (from DEBT_CD.FILING_ID)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing',
            'line_item',
        ),)
        verbose_name = "Form 460 (Campaign Disclosure) Schedule F item"

    def __str__(self):
        return '%s-%s' % (self.filing, self.line_item)


class Form460ScheduleFItemVersion(Form460ScheduleFItemBase):
    """
    Every version of each accrued expense of a campaign filer.

    For accrued expenses itemized on Schedule F of the most recent version of
    each Form 460 filing, see Form460ScheduleGItem.

    Derived from DEBT_CD records.
    """
    filing_version = models.ForeignKey(
        'Form460FilingVersion',
        related_name='schedule_f_items',
        null=True,
        on_delete=models.SET_NULL,
        help_text='Foreign key referring to the version of the Form 460 that '
                  'includes the Schedule F items'
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing_version',
            'line_item',
        ),)
        index_together = ((
            'filing_version',
            'line_item',
        ),)
        verbose_name = "Form 460 (Campaign Disclosure) Schedule F item version"

    def __str__(self):
        return '%s-%s-%s' % (
            self.filing_version.filing_id,
            self.filing_version.amend_id,
            self.line_item
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom manager for the OCD Organization model.
"""
from __future__ import unicode_literals
from calaccess_raw.models import FilerToFilerTypeCd
from calaccess_processed.managers import BulkLoadSQLManager


class OCDPartyManager(BulkLoadSQLManager):
    """
    Limited the OCD Organization model to politics parties.
    """
    def get_queryset(self):
        """
        Override the default manager to limit the results to political parties.
        """
        return super(OCDPartyManager, self).get_queryset().filter(classification='party')

    def unknown(self):
        """
        Returns the UNKNOWN party.
        """
        return self.get_queryset().get(name='UNKNOWN')

    def get_by_name(self, name):
        """
        Helper for getting the OCD party object giving a raw name from CAL-ACCESS.

        If not found, return the "UNKNOWN" Organization object.
        """
        # First try a full name
        try:
            return self.get_queryset().get(name=name)
        except self.model.DoesNotExist:
            pass

        # If that doesn't work, try an alternate name
        try:
            return self.get_queryset().get(other_names__name=name)
        except self.model.DoesNotExist:
            pass

        # And if that doesn't work, just return the unknown party object
        return self.unknown()

    def get_by_filer_id(self, filer_id, election_date):
        """
        Lookup the party for the given filer_id, effective before election_date.

        If not found, return the "UNKNOWN" Organization object.
        """
        # Try to see if the record exists in the raw data with a party code
        try:
            party_code = FilerToFilerTypeCd.objects.filter(
                filer_id=filer_id,
                effect_dt__lte=election_date,
            ).latest('effect_dt').party_cd
        except FilerToFilerTypeCd.DoesNotExist:
            # If it doesn't hit just quit now
            return self.unknown()

        # IF we have a code, transform "INDEPENDENT" and "NON-PARTISAN" codes to "NO PARTY PREFERENCE"
        if party_code in [16007, 16009, 0]:
            party_code = 16012

        # Try pulling out the party using the lookup code
        try:
            return self.get_queryset().get(identifiers__identifier=party_code)
        except self.model.DoesNotExist:
            pass

        # If that fails, just quit and return the unknown party object
        return self.unknown()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom managers for the OCD Post model.
"""
from __future__ import unicode_literals
import re
from django.db.models import Q
from calaccess_processed.managers import BulkLoadSQLManager


class OCDPostManager(BulkLoadSQLManager):
    """
    Custom helpers for the OCD Post model.
    """
    def parse_office_name(self, office_name):
        """
        Parse string containg the name for an office.

        Expected format is "{TYPE NAME} [{DISTRICT NUMBER}]".

        Return a dict with two keys: type and district.
        """
        office_pattern = r'^(?P<type>[A-Z ]+)(?P<district>\d{2})?$'
        try:
            parsed = re.match(office_pattern, office_name.upper()).groupdict()
        except AttributeError:
            parsed = {'type': None, 'district': None}
        else:
            parsed['type'] = parsed['type'].strip()
            try:
                parsed['district'] = int(parsed['district'])
            except TypeError:
                pass
        return parsed

    def get_by_name(self, office_name, method="get"):
        """
        Get a Post object with an office string.
        """
        from calaccess_processed_elections.proxies import OCDDivisionProxy, OCDOrganizationProxy

        parsed_office = self.parse_office_name(office_name)

        # prepare to get or create post
        label = office_name.title().replace('Of', 'of')

        if parsed_office['type'] == 'STATE SENATE':
            division = OCDDivisionProxy.senate.get(subid2=parsed_office['district'])
            organization = OCDOrganizationProxy.objects.senate()
            role = 'Senator'
        elif parsed_office['type'] == 'ASSEMBLY':
            division = OCDDivisionProxy.assembly.get(subid2=parsed_office['district'])
            organization = OCDOrganizationProxy.objects.assembly()
            role = 'Assembly Member'
        else:
            # If not Senate or Assembly, assume this is a state office
            division = OCDDivisionProxy.objects.california()
            if parsed_office['type'] == 'MEMBER BOARD OF EQUALIZATION':
                organization = OCDOrganizationProxy.objects.board_of_equalization()
                role = 'Board Member'
            elif parsed_office['type'] == 'SECRETARY OF STATE':
                organization = OCDOrganizationProxy.objects.secretary_of_state()
                role = label
            else:
                organization = OCDOrganizationProxy.objects.executive_branch()
                role = label

        # Grab the method passed in. You can see why we did this in the method just below this one.
        func = getattr(self.get_queryset(), method)

        # Run it pass back the result.
        return func(
            label=label,
            role=role,
            division=division,
            organization=organization
        )

    def get_or_create_by_name(self, office_name):
        """
        Get or create a Post object with an office_name string.

        Returns a tuple (Post object, created), where created is a boolean specifying whether a Post was created.
        """
        # We'll use a hack on the method above to get this done so we can avoid repeating code.
        return self.get_by_name(office_name, method="get_or_create")

    def get_by_form501(self, form501):
        """
        Get a Post using data extracted from Form501Filing.

        Return Post object or None if not found.
        """
        from calaccess_processed_elections.proxies import RawFilerToFilerTypeCdProxy, OCDDivisionProxy

        # Try to get it using office_name
        try:
            return self.get_by_name(form501.office_name)
        except (self.model.DoesNotExist, OCDDivisionProxy.DoesNotExist):
            pass

        # Try extracting office and district from FilerToFilerTypeCd
        filer_id_office_name = RawFilerToFilerTypeCdProxy.objects.get_office_by_filer_id_and_date(
            form501.filer_id,
            form501.ocd_election.date
        )

        # If you can't find that, just quit
        if not filer_id_office_name:
            return None

        # If you can, try to get that
        try:
            return self.get_by_name(filer_id_office_name)
        except (self.model.DoesNotExist, OCDDivisionProxy.DoesNotExist):
            return None


class OCDAssemblyPostManager(BulkLoadSQLManager):
    """
    Custom manager for State Assembly office Posts.
    """
    def get_queryset(self):
        """
        Filters down to State Assembly posts.
        """
        return super(OCDAssemblyPostManager, self).get_queryset().filter(
            organization__name='California State Assembly',
        )


class OCDExecutivePostManager(BulkLoadSQLManager):
    """
    Custom manager for State Executive Branch office Posts.
    """
    def get_queryset(self):
        """
        Filters down to State Executive Branch posts.
        """
        return super(OCDExecutivePostManager, self).get_queryset().filter(
            Q(organization__name='California State Executive Branch')
            | Q(organization__parent__name='California State Executive Branch')
        )


class OCDSenatePostManager(BulkLoadSQLManager):
    """
    Custom manager for State Senate office Posts.
    """
    def get_queryset(self):
        """
        Filters down to State Senate posts.
        """
        return super(OCDSenatePostManager, self).get_queryset().filter(
            organization__name='California State Senate',
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Unittests Filing and FilingVersion models.
"""
from django.test import TestCase
import requests_mock
from calaccess_processed_filings.models import Form460Filing


class Form460FilingTest(TestCase):
    """
    Test for Form460Filing model.
    """
    fixtures = ['form460filing.json', 'form460filingversion.json']

    def test_filing_has_pdf(self):
        """
        Confirm has_pdf returns True for the filing instance.
        """
        filing = Form460Filing.objects.all()[0]
        with requests_mock.Mocker() as m:
            m.register_uri('HEAD', filing.pdf_url)
            self.assertIs(filing.has_pdf, True)

    def test_filing_version_has_pdf(self):
        """
        Confirm has_pdf returns True for the filing instance.
        """
        filing_version = Form460Filing.objects.all()[0].versions.all()[0]
        with requests_mock.Mocker() as m:
            m.register_uri('HEAD', filing_version.pdf_url)
            self.assertIs(filing_version.has_pdf, True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Abstract base models for campaign finance-related filings and transactions.
"""
from django.db import models
from calaccess_processed_filings.models.base import FilingBaseModel


class CampaignFinanceFilingBase(FilingBaseModel):
    """
    Base and abstract model for campaign finance-related filings.
    """
    date_filed = models.DateField(
        verbose_name='date filed',
        db_index=True,
        null=False,
        help_text="Date this report was filed, according to the filer "
                  "(from CVR_CAMPAIGN_DISCLOSURE.RPT_DATE)",
    )
    filer_id = models.IntegerField(
        verbose_name='filer id',
        db_index=True,
        null=False,
        help_text="Numeric filer identification number (from FILER_XREF.FILER_ID)",
    )
    filer_lastname = models.CharField(
        verbose_name='filer last name',
        max_length=200,
        null=False,
        blank=False,
        help_text="Last name of filer (from CVR_CAMPAIGN_DISCLOSURE.FILER_NAML)",
    )
    filer_firstname = models.CharField(
        verbose_name="filer first name",
        max_length=45,
        null=False,
        blank=True,
        help_text="First name of the filer (from "
                  "CVR_CAMPAIGN_DISCLOSURE.FILER_NAMF)",
    )
    election_date = models.DateField(
        verbose_name='election date',
        db_index=True,
        null=True,
        help_text="Date of the election in which the filer is participating "
                  "(from CVR_CAMPAIGN_DISCLOSURE.ELECT_DATE)",
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed'
        abstract = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Load OCD BallotMeasureContest and related models with scraped CAL-ACCESS data.
"""
from opencivicdata.elections.models import BallotMeasureContest
from calaccess_processed.management.commands import CalAccessCommand
from calaccess_processed_elections.proxies import ScrapedPropositionProxy, OCDDivisionProxy


class Command(CalAccessCommand):
    """
    Load OCD BallotMeasureContest and related models with scraped CAL-ACCESS data.
    """
    help = 'Load OCD BallotMeasureContest and related models with scraped CAL-ACCESS data'

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(Command, self).handle(*args, **options)
        self.header('Loading Ballot Measure Contests')
        self.load()
        self.success("Done!")

    def create_contest(self, scraped_prop, ocd_elec):
        """
        Create an OCD BallotMeasureContest object derived from a ScrapedProposition.

        Return a BallotMeasureContest object.
        """
        # Create the object
        return BallotMeasureContest.objects.create(
            election=ocd_elec,
            division=OCDDivisionProxy.objects.california(),
            name=scraped_prop.name,
            classification=scraped_prop.classification,
        )

    def load(self):
        """
        Load OCD ballot measure-related models with data scraped from CAL-ACCESS website.
        """
        object_list = ScrapedPropositionProxy.objects.exclude(name__icontains='RECALL')
        for scraped_prop in object_list:
            ocd_election = scraped_prop.election_proxy.get_ocd_election()
            try:
                # Try getting the contest using scraped_id
                ocd_contest = ocd_election.ballotmeasurecontests.get(
                    identifiers__scheme='calaccess_measure_id',
                    identifiers__identifier=scraped_prop.scraped_id,
                )
            except BallotMeasureContest.DoesNotExist:
                # If not there, create one
                ocd_contest = self.create_contest(scraped_prop, ocd_election)
                # Add the options
                ocd_contest.options.create(text='yes')
                ocd_contest.options.create(text='no')
                # Add the identifiers
                ocd_contest.identifiers.create(
                    scheme='calaccess_measure_id',
                    identifier=scraped_prop.scraped_id,
                )
                if self.verbosity > 2:
                    self.log(
                        'Created {0}: {1}'.format(
                            ocd_contest._meta.object_name,
                            ocd_contest,
                        )
                    )
            else:
                # If the contest already exists, make sure the name is up-to-date
                # because they could change on subsequent scrapes of the SoS website
                if ocd_contest.name != scraped_prop.name:
                    ocd_contest.name = scraped_prop.name
                    ocd_contest.save()

            # Update or create the Contest source
            ocd_contest.sources.update_or_create(
                url=scraped_prop.url,
                note='Last scraped on {dt:%Y-%m-%d}'.format(
                    dt=scraped_prop.last_modified,
                )
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages
from distutils.core import Command


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


def version_scheme(version):
    """
    Version scheme hack for setuptools_scm.
    Appears to be necessary to due to the bug documented here: https://github.com/pypa/setuptools_scm/issues/342
    If that issue is resolved, this method can be removed.
    """
    import time

    from setuptools_scm.version import guess_next_version

    if version.exact:
        return version.format_with("{tag}")
    else:
        _super_value = version.format_next_version(guess_next_version)
        now = int(time.time())
        return _super_value + str(now)


def local_version(version):
    """
    Local version scheme hack for setuptools_scm.
    Appears to be necessary to due to the bug documented here: https://github.com/pypa/setuptools_scm/issues/342
    If that issue is resolved, this method can be removed.
    """
    return ""


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from django.conf import settings
        settings.configure(
            DATABASES={
                'default': {
                    'NAME': ':memory:',
                    'ENGINE': 'django.db.backends.sqlite3'
                }
            },
            INSTALLED_APPS=('calaccess_processed',),
            MIDDLEWARE_CLASSES=()
        )
        from django.core.management import call_command
        import django
        django.setup()
        call_command('test', 'calaccess_processed')


setup(
    name='django-calaccess-processed-data',
    license='MIT',
    description='A Django app to transform and refine campaign finance data from the California Secretary of State’s \
CAL-ACCESS database',
    long_description=read('README.rst'),
    url='http://django-calaccess.californiacivicdata.org',
    author='California Civic Data Coalition',
    author_email='b@palewi.re',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,  # because we're including static files
    install_requires=(
        'django-calaccess-raw-data==4.1.*',
        'django-calaccess-scraped-data==3.1.0',
        'django>=3.2.*',
        'ccdc-opencivicdata==0.0.1',
        'django-internetarchive-storage',
        'pytz',
    ),
    cmdclass={'test': TestCommand,},
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'License :: OSI Approved :: MIT License',
    ),
    setup_requires=["setuptools_scm"],
    use_scm_version={"version_scheme": version_scheme, "local_scheme": local_version},
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Export and archive a .csv file for a given model.
"""
from django.apps import apps
from calaccess_processed.management.commands._archivecalaccessprocessedfile import Command as BaseCommand


class Command(BaseCommand):
    """
    A version of the parent command for this app.
    """
    def get_model(self, processed_file):
        """
        Submit a processed_file and receive the proper proxy.
        """
        model_list = apps.get_app_config("calaccess_processed_filings").get_filing_models()
        model_dict = dict((m.__name__, m) for m in model_list)
        return model_dict[processed_file.file_name]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Abstract base models for Form 496 models.
"""
from .filing import Form496Filing, Form496FilingVersion
from .part1 import (
    Form496Part1Item,
    Form496Part1ItemVersion
)
from .part2 import (
    Form496Part2Item,
    Form496Part2ItemVersion
)
from .part3 import (
    Form496Part3Item,
    Form496Part3ItemVersion
)

__all__ = (
    "Form496Filing",
    "Form496FilingVersion",
    "Form496Part1Item",
    "Form496Part1ItemVersion",
    "Form496Part2Item",
    "Form496Part2ItemVersion",
    "Form496Part3Item",
    "Form496Part3ItemVersion"
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Models for storing data from Schedule 497, the Late Contribution Reports.

More about the filing: http://calaccess.californiacivicdata.org/documentation/calaccess-forms/f497/
"""
from django.db import models
from .base import Form497ItemBase


class Form497Part1ItemBase(Form497ItemBase):
    """
    Abstract base model for items reported on Part 1 of Schedule 497 filings.

    On Part 1 of Schedule 497, campaign filers are required to report
    contributions received in the 90 days leading up to an election.
    """
    CONTRIBUTOR_CODE_CHOICES = (
        ('BNM', 'Ballot measure name/title'),
        ('CAO', 'Candidate/officeholder'),
        ('COM', 'Committee'),
        ('CTL', 'Controlled committee'),
        ('IND', 'Individual'),
        ('OFF', 'Officer'),
        ('OTH', 'Other'),
        ('PTY', 'Political Party'),
        ('RCP', 'Recipient committee'),
        ('SCC', 'Small Contributor Committee'),
    )
    contributor_code = models.CharField(
        verbose_name='contributor code',
        max_length=3,
        blank=True,
        choices=CONTRIBUTOR_CODE_CHOICES,
        help_text='Code describing the contributor (from S497_CD.ENTITY_CD)',
    )
    contributor_committee_id = models.CharField(
        verbose_name='committee id',
        max_length=9,
        blank=True,
        help_text="Contributor's filer identification number, if it is a "
                  "committee (from RCPT_CD.CMTE_ID)",
    )
    contributor_title = models.CharField(
        verbose_name='contributor title',
        max_length=10,
        blank=True,
        help_text='Name title of the contributor (from S497_CD.ENTY_NAMT)',
    )
    contributor_lastname = models.CharField(
        verbose_name='contributor lastname',
        max_length=200,
        blank=True,
        help_text='Last name of the contributor (from S497_CD.ENTY_NAML)',
    )
    contributor_firstname = models.CharField(
        verbose_name='contributor firstname',
        max_length=45,
        help_text='First name of the contributor (from S497_CD.ENTY_NAMF)',
    )
    contributor_name_suffix = models.CharField(
        verbose_name='contributor name suffix',
        max_length=10,
        blank=True,
        help_text='Name suffix of the contributor (from S497_CD.ENTY_NAMS)',
    )
    contributor_city = models.CharField(
        verbose_name='contributor city',
        max_length=30,
        blank=True,
        help_text='City of the contributor (from S497_CD.ENTY_CITY)',
    )
    contributor_state = models.CharField(
        verbose_name='contributor state',
        max_length=2,
        blank=True,
        help_text='State of the contributor (from S497_CD.ENTY_ST)',
    )
    contributor_zip = models.CharField(
        verbose_name='contributor zip',
        max_length=10,
        blank=True,
        help_text='Zip code (usually zip5, sometimes zip9) of the '
                  'contributor (from S497_CD.ENTY_ZIP4)',
    )
    contributor_employer = models.CharField(
        verbose_name='contributor employer',
        max_length=200,
        blank=True,
        help_text='Employer of the contributor (from S497_CD.CTRIB_EMP)',
    )
    contributor_occupation = models.CharField(
        verbose_name='contributor occupation',
        max_length=60,
        blank=True,
        help_text='Occupation of the contributor (from S497_CD.CTRIB_OCC)',
    )
    contributor_is_self_employed = models.BooleanField(
        verbose_name='contributor is self employed',
        default=False,
        help_text='Indicates whether or not the contributor is self-employed'
                  '(from S497_CD.CTRIB_SELF)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        abstract = True


class Form497Part1Item(Form497Part1ItemBase):
    """
    Late contributions received by campaign filers.

    These transactions are itemized on Part 1 of the most recent version
    of each Schedule 497 filing. For receipts of late contributions itemized
    on any version of any Schedule 497 filing, see Form497Part1ItemVersion.

    Derived from S497_CD records where FORM_TYPE is 'F497P1'.
    """
    filing = models.ForeignKey(
        'Form497Filing',
        related_name='contributions_received',
        null=True,
        on_delete=models.SET_NULL,
        help_text='Unique identification number for the Schedule 497 filing ('
                  'from S497_CD.FILING_ID)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing',
            'line_item',
        ),)
        verbose_name = "Form 497 (Late Contribution) Part 1 item"

    def __str__(self):
        return '%s-%s' % (self.filing, self.line_item)


class Form497Part1ItemVersion(Form497Part1ItemBase):
    """
    Every version of each late contribution received by a campaign filer.

    For late contributions itemized on Part 1 of the most recent version of
    each Schedule 497 filing, see Form497Part1Item.

    Derived from S497_CD records where FORM_TYPE is 'F497P1'.
    """
    filing_version = models.ForeignKey(
        'Form497FilingVersion',
        related_name='contributions_received',
        null=True,
        on_delete=models.SET_NULL,
        help_text='Foreign key referring to the version of the Schedule 497 '
                  'that includes the received contribution'
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing_version',
            'line_item',
        ),)
        index_together = ((
            'filing_version',
            'line_item',
        ),)
        verbose_name = "Form 497 (Late Contribution) Part 1 item version"

    def __str__(self):
        return '%s-%s-%s' % (
            self.filing_version.filing_id,
            self.filing_version.amend_id,
            self.line_item
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Managers for generating flatfiles that combine multiple table into a simplified file.
"""
from __future__ import unicode_literals
from django.db.models import Q, F, Max
from calaccess_processed.managers import BulkLoadSQLManager


class OCDFlatBallotMeasureContestManager(BulkLoadSQLManager):
    """
    Custom manager for flattening contents of the OCD BallotMeasureContest model.
    """
    def get_queryset(self):
        """
        Returns the custom QuerySet for this manager.
        """
        qs = super(
            OCDFlatBallotMeasureContestManager, self
        ).get_queryset()
        return qs.filter(
            Q(identifiers__scheme='calaccess_measure_id')
            | Q(identifiers__isnull=True)
        ).annotate(
            election_name=F('election__name'),
            election_date=F('election__date'),
            ocd_contest_id=F('id'),
            ocd_election_id=F('election_id'),
            calaccess_measure_id=Max('identifiers__identifier')
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Base class for proxies to OCD models.
"""
import textwrap
from .fields import CopyToField
from django.template.defaultfilters import capfirst


class OCDProxyModelMixin(object):
    """
    Properties and methods shared by all OCD proxy models.
    """
    @property
    def base_model(self):
        """
        The model being proxied.
        """
        return self.__class__.__bases__[0]

    @property
    def meta(self):
        """
        The model's meta attributes.
        """
        return self._meta

    @property
    def is_flat(self):
        """
        True if the proxy model is used to flatten relational data models.
        """
        return 'Flat' in self._meta.object_name

    @property
    def file_name(self):
        """
        The name for the csv to which the model's contents will be dumped.

        If the model is a flat model proxy, return the model's verbose_name_plural
        in CamelCase. Otherwise, return the object_name of the base_model.
        """
        if self.is_flat:
            file_name = ''.join(
                x for x in str(self._meta.verbose_name_plural).title()
                if not x.isspace()
            )
        else:
            file_name = self.base_model._meta.object_name
        return file_name

    @property
    def display_name(self):
        """
        The version of the name for this model we prefer to display on public facing web pages.
        """
        if self.is_flat:
            return capfirst(self.meta.verbose_name_plural)
        else:
            return self.file_name

    @property
    def doc(self):
        """
        Doc string of the proxy model, or base_model (if not flat).
        """
        if self.is_flat:
            doc = textwrap.dedent(self.__doc__).strip()
        elif self.base_model.__doc__.startswith(self.base_model._meta.object_name):
            doc = ''
        else:
            doc = textwrap.dedent(self.base_model.__doc__).strip()
        return doc

    @property
    def klass_group(self):
        """
        The model's group.
        """
        if self.is_flat:
            group = "Flat"
        else:
            group = "Relational"
        return group

    def get_field_list(self):
        """
        Return all the fields on the model as a tuple.
        """
        try:
            self.copy_to_fields
        except AttributeError:
            fields = self._meta.fields
        else:
            q = self._meta.model.objects.all().query
            fields_list = []
            for f in self.copy_to_fields:
                if len(f) > 1:
                    field = CopyToField(q, f[0], help_text=f[1])
                else:
                    field = CopyToField(q, f[0])
                fields_list.append(field)
            fields = tuple(f for f in fields_list)
        return fields



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Proxy models for augmenting our source data tables with methods useful for processing.
"""
from __future__ import unicode_literals
from django.db.models import Q
from postgres_copy import CopyQuerySet
from calaccess_processed.managers import BulkLoadSQLManager


class OCDCandidacyQuerySet(CopyQuerySet):
    """
    Custom QuerySet for the OCD Candidacy model.
    """
    def get_by_filer_id(self, filer_id):
        """
        Returns a Candidacy object linked to a CALACCESS filer_id, if it exists.
        """
        return self.get(
            person__identifiers__scheme='calaccess_filer_id',
            person__identifiers__identifier=filer_id,
        )

    def get_by_name(self, name):
        """
        Returns a Candidacy object with the provided name from the CALACCESS database or scrape.
        """
        q = self.filter(
            Q(candidate_name=name)
            | Q(person__name=name)
            | Q(person__other_names__name=name)
        ).distinct()

        if not q.exists():
            raise self.model.DoesNotExist('OCDCandidacyProxy matching query does not exist')
        elif q.count() > 1:
            raise self.model.MultipleObjectsReturned(
                'get_by_name() returned more than one OCDCandidacyProxy -- it returned %s!' % q.count()
            )
        else:
            return q[0]


class OCDCandidacyManager(BulkLoadSQLManager):
    """
    Manager for custom methods on the OCDCandidacyProxy model.
    """
    def get_queryset(self):
        """
        Returns the custom QuerySet for this manager.
        """
        return OCDCandidacyQuerySet(self.model, using=self._db)

    def matched_form501_ids(self):
        """
        Return all the Form 501 filing ids matched to a candidacy record.
        """
        return [
            i['extras']['form501_filing_ids'] for i in
            self.get_queryset().filter(extras__has_key='form501_filing_ids').values('extras')
        ]

    def get_or_create_from_calaccess(
        self,
        contest,
        candidate_name_dict,
        candidate_status="filed",
        candidate_filer_id=None
    ):
        """
        Get or create a Candidacy object with data from the CAL-ACCESS database.

        First, try getting an existing Candidacy within the given CandidateContest
        linked to a Person with the provided filer_id. If matched and the matched Person
        has different current name and doesn't have the provided name as an other name,
        add the other name.

        Next, try getting an existing Candidacy within the given CandidateContest
        linked to a Person with provided name (as default name or other name). If
        matched and match candidate doesn't already have filer_id, add the filer_id.

        If no match or if the matched person already has a different filer_id, create
        a new Candidacy (this may also create a new Person record).

        Returns a tuple (Candidacy object, created), where created is a boolean
        specifying whether a Candidacy was created.
        """
        from calaccess_processed_elections.proxies import OCDPersonProxy, OCDCandidacyProxy

        candidacy = None

        # first, try matching to existing candidate in contest with filer_id
        if candidate_filer_id:
            try:
                candidacy = self.model.objects.filter(contest=contest).get_by_filer_id(candidate_filer_id)
            except self.model.DoesNotExist:
                pass
            else:
                candidacy_created = False
                # if provided name not person's current name and not linked to person add it
                candidacy.person_proxy.add_other_name(
                    candidate_name_dict['name'],
                    'Matched on CandidateContest and calaccess_filer_id'
                )

        # if filer_id match fails (or no filer_id), try matching to candidate
        # in contest with provided name
        if not candidacy:
            try:
                candidacy = self.model.objects.filter(contest=contest).get_by_name(candidate_name_dict['name'])
            except self.model.DoesNotExist:
                pass
            else:
                candidacy_created = False
                # if filer_id provided
                if candidate_filer_id:
                    # check to make sure candidate with same name doesn't have diff filer_id
                    if candidacy.person.identifiers.filter(scheme='calaccess_filer_id').exists():
                        # if so, don't conflate
                        candidacy = None
                    else:
                        # if so, add filer_id to existing candidate
                        person = candidacy.person
                        person.refresh_from_db()
                        person.__class__ = OCDPersonProxy
                        person.add_filer_id(candidate_filer_id)

        # if no matched candidate yet, make a new one
        if not candidacy:
            # First make a Person object
            person, person_created = OCDPersonProxy.objects.get_or_create_from_calaccess(
                candidate_name_dict,
                candidate_filer_id=candidate_filer_id
            )
            person.add_other_name(candidate_name_dict['name'], 'From {} candidacy'.format(contest))

            # Then make the Candidacy
            candidacy = OCDCandidacyProxy.objects.create(
                contest=contest,
                person=person,
                post=contest.posts.all()[0].post,
                candidate_name=candidate_name_dict['name'],
                registration_status=candidate_status,
            )
            candidacy_created = True

        # if provided registration does not equal the default, update
        if candidate_status != 'filed' and candidate_status != candidacy.registration_status:
            candidacy.registration_status = candidate_status
            candidacy.save()

        # make sure Person name is same as most recent candidate_name
        person = candidacy.person
        person.refresh_from_db()
        person.__class__ = OCDPersonProxy
        person.update_name()

        # Pass it back out.
        return candidacy, candidacy_created



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Proxy models for generating flatfiles that combine multiple table into a simplified file.
"""
from opencivicdata.elections.models import BallotMeasureContest
from calaccess_processed.proxies import OCDProxyModelMixin
from calaccess_processed_flatfiles.managers import OCDFlatBallotMeasureContestManager


class OCDFlatBallotMeasureContestProxy(BallotMeasureContest, OCDProxyModelMixin):
    """
    Every ballot measure.
    """
    objects = OCDFlatBallotMeasureContestManager()

    copy_to_fields = (
        ('name', 'Name of the ballot measure, not necessarily as it appears on the ballot.'),
        ('classification',),
        ('election_name', 'Name of the election in which the ballot measure is decided.'),
        ('election_date', 'Date of the election in which the ballot measure is decided.'),
        ('description',),
        ('created_at',),
        ('updated_at',),
        ('ocd_contest_id',),
        ('ocd_election_id', BallotMeasureContest._meta.get_field('election').help_text),
        ('calaccess_measure_id', 'Identifier assigned to the ballot measure by CAL-ACCESS.'),
    )

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_flatfiles"
        proxy = True
        verbose_name_plural = 'ballot measures'



# Generated by Django 3.2.4 on 2021-06-11 13:24

import calaccess_processed.proxies
from django.db import migrations


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('elections', '0008_auto_20181029_1527'),
    ]

    operations = [
        migrations.CreateModel(
            name='OCDFlatBallotMeasureContestProxy',
            fields=[
            ],
            options={
                'verbose_name_plural': 'ballot measures',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('elections.ballotmeasurecontest', calaccess_processed.proxies.OCDProxyModelMixin),
        ),
        migrations.CreateModel(
            name='OCDFlatCandidacyProxy',
            fields=[
            ],
            options={
                'verbose_name_plural': 'candidates',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('elections.candidacy', calaccess_processed.proxies.OCDProxyModelMixin),
        ),
        migrations.CreateModel(
            name='OCDFlatRetentionContestProxy',
            fields=[
            ],
            options={
                'verbose_name_plural': 'recall measures',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('elections.retentioncontest', calaccess_processed.proxies.OCDProxyModelMixin),
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Proxy models for augmenting our source data tables with methods useful for processing.
"""
from __future__ import unicode_literals
from opencivicdata.elections.models import (
    RetentionContest,
    RetentionContestIdentifier,
    RetentionContestOption,
    RetentionContestSource
)
from calaccess_processed.managers import BulkLoadSQLManager
from calaccess_processed.proxies import OCDProxyModelMixin


class OCDRetentionContestProxy(RetentionContest, OCDProxyModelMixin):
    """
    A proxy on the OCD RetentionContest model.
    """
    objects = BulkLoadSQLManager()

    copy_to_fields = (
        ('id',),
        ('name',),
        ('membership_id',),
        ('division_id',),
        ('election_id',),
        ('description',),
        ('requirement',),
        ('runoff_for_contest_id',),
        ('created_at',),
        ('updated_at',),
        ('extras',),
        ('locked_fields',),
    )

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True


class OCDRetentionContestIdentifierProxy(RetentionContestIdentifier, OCDProxyModelMixin):
    """
    A proxy on the OCD RetentionContestIdentifier model.
    """
    objects = BulkLoadSQLManager()

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True


class OCDRetentionContestOptionProxy(RetentionContestOption, OCDProxyModelMixin):
    """
    A proxy on the OCD RetentionContestOption model.
    """
    objects = BulkLoadSQLManager()

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True


class OCDRetentionContestSourceProxy(RetentionContestSource, OCDProxyModelMixin):
    """
    A proxy on the OCD RetentionContestSource model.
    """
    objects = BulkLoadSQLManager()

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Basic configuration for the application.
"""
from __future__ import unicode_literals, absolute_import
import os
import collections
from django.apps import apps
from django.apps import AppConfig


class CalAccessProcessedElectionsConfig(AppConfig):
    """
    Application configuration.
    """
    name = 'calaccess_processed_elections'
    verbose_name = "CAL-ACCESS processed data: Elections"
    default_auto_field = 'django.db.models.AutoField'
    # Where SQL files are stored in this application
    sql_directory_path = os.path.join(os.path.dirname(__file__), 'sql')

    def get_ocd_models_list(self):
        """
        Returns a list of all the OCD models proxied by this app.
        """
        return list(self.get_ocd_models_map().keys())

    def get_ocd_proxy_lookup(self):
        """
        Returns a dictionary with the names of data models mapped to proxies.
        """
        # Convert the keys to strings
        return dict((k.__name__, v) for k, v in self.get_ocd_models_map().items())

    def get_ocd_models_map(self):
        """
        Returns a list of the models that should be saved in our archive.
        """
        from . import proxies
        ocd_core = apps.get_app_config('core')
        ocd_elections = apps.get_app_config('elections')

        # Create a dict mapping the models to proxies
        return collections.OrderedDict({
            ocd_core.get_model('Division'): proxies.OCDDivisionProxy,
            ocd_core.get_model('Organization'): proxies.OCDOrganizationProxy,
            ocd_core.get_model('OrganizationIdentifier'): proxies.OCDOrganizationIdentifierProxy,
            ocd_core.get_model('OrganizationName'): proxies.OCDOrganizationNameProxy,
            ocd_core.get_model('Jurisdiction'): proxies.OCDJurisdictionProxy,
            ocd_core.get_model('Post'): proxies.OCDPostProxy,
            ocd_core.get_model('Person'): proxies.OCDPersonProxy,
            ocd_core.get_model('PersonIdentifier'): proxies.OCDPersonIdentifierProxy,
            ocd_core.get_model('PersonName'): proxies.OCDPersonNameProxy,
            ocd_core.get_model('Membership'): proxies.OCDMembershipProxy,
            ocd_elections.get_model('Election'): proxies.OCDElectionProxy,
            ocd_elections.get_model('ElectionIdentifier'): proxies.OCDElectionIdentifierProxy,
            ocd_elections.get_model('ElectionSource'): proxies.OCDElectionSourceProxy,
            ocd_elections.get_model('Candidacy'): proxies.OCDCandidacyProxy,
            ocd_elections.get_model('CandidacySource'): proxies.OCDCandidacySourceProxy,
            ocd_elections.get_model('BallotMeasureContest'): proxies.OCDBallotMeasureContestProxy,
            ocd_elections.get_model('BallotMeasureContestOption'): proxies.OCDBallotMeasureContestOptionProxy,
            ocd_elections.get_model('BallotMeasureContestIdentifier'): proxies.OCDBallotMeasureContestIdentifierProxy,
            ocd_elections.get_model('BallotMeasureContestSource'): proxies.OCDBallotMeasureContestSourceProxy,
            ocd_elections.get_model('RetentionContest'): proxies.OCDRetentionContestProxy,
            ocd_elections.get_model('RetentionContestOption'): proxies.OCDRetentionContestOptionProxy,
            ocd_elections.get_model('RetentionContestIdentifier'): proxies.OCDRetentionContestIdentifierProxy,
            ocd_elections.get_model('RetentionContestSource'): proxies.OCDRetentionContestSourceProxy,
            ocd_elections.get_model('CandidateContest'): proxies.OCDCandidateContestProxy,
            ocd_elections.get_model('CandidateContestPost'): proxies.OCDCandidateContestPostProxy,
            ocd_elections.get_model('CandidateContestSource'): proxies.OCDCandidateContestSourceProxy
        })



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Proxy models for augmenting our source data tables with methods useful for processing.
"""
from __future__ import unicode_literals
from opencivicdata.elections.models import (
    BallotMeasureContest,
    BallotMeasureContestIdentifier,
    BallotMeasureContestOption,
    BallotMeasureContestSource,
)
from calaccess_processed.proxies import OCDProxyModelMixin
from calaccess_processed.managers import BulkLoadSQLManager


class OCDBallotMeasureContestProxy(BallotMeasureContest, OCDProxyModelMixin):
    """
    A proxy on the OCD BallotMeasureContest model.
    """
    objects = BulkLoadSQLManager()

    copy_to_fields = (
        ('id',),
        ('name',),
        ('division_id',),
        ('election_id',),
        ('description',),
        ('requirement',),
        ('classification',),
        ('runoff_for_contest_id',),
        ('created_at',),
        ('updated_at',),
        ('extras',),
        ('locked_fields',),
    )

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True


class OCDBallotMeasureContestIdentifierProxy(BallotMeasureContestIdentifier, OCDProxyModelMixin):
    """
    A proxy on the OCD BallotMeasureContestIdentifier model.
    """
    objects = BulkLoadSQLManager()

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True


class OCDBallotMeasureContestOptionProxy(BallotMeasureContestOption, OCDProxyModelMixin):
    """
    A proxy on the OCD BallotMeasureContestOption model.
    """
    objects = BulkLoadSQLManager()

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True


class OCDBallotMeasureContestSourceProxy(BallotMeasureContestSource, OCDProxyModelMixin):
    """
    A proxy on the OCD BallotMeasureContestSource model.
    """
    objects = BulkLoadSQLManager()

    class Meta:
        """
        Make this a proxy model.
        """
        app_label = "calaccess_processed_elections"
        proxy = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Load data from CAL-ACCESS' Form 460 into standardized OCD models.
"""
from calaccess_processed import models
from calaccess_processed.management.commands import CalAccessCommand


class Command(CalAccessCommand):
    """
    Load data from CAL-ACCESS' Form 460 into standardized OCD models.
    """
    help = "Load data from CAL-ACCESS' Form 460 into standardized OCD models."

    def add_arguments(self, parser):
        """
        Adds custom arguments specific to this command.
        """
        parser.add_argument(
            "--flush",
            action="store_true",
            dest="flush",
            default=False,
            help="Flush OCD filing models prior to loading"
        )

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(Command, self).handle(*args, **options)
        if options['flush']:
            self.flush()
        self.load()

    def flush(self, *args, **kwargs):
        """
        Clear out the OCD models loaded by this command.
        """
        self.header('Flushing OCD data extracted from Form 460 filings')
        model_list = [
            # models.OCDTransactionProxy,
            models.OCDFilingActionSummaryAmountProxy,
            models.OCDFilingActionProxy,
            models.OCDFilingProxy,
            models.OCDFilingIdentifierProxy,
            models.OCDCommitteeProxy,
            models.OCDCommitteeIdentifierProxy,
            models.OCDCommitteeNameProxy,
            models.OCDCommitteeTypeProxy,
        ]
        for m in model_list:
            self.log(" Flushing {}".format(m.__name__))
            m.objects.all().delete()

    def load_model(self, model):
        """
        Loads the provided model.
        """
        # Print out what we're doing
        self.log('  Loading {}'.format(model.__name__))

        # Count how many records are in the table now
        before = model.objects.count()

        # Run the update routine
        model.objects.load()

        # Count how many records are in the table afterward
        after = model.objects.count()

        # Report the change
        self.log('   {:,} added'.format(after - before))

    def load(self, *args, **kwargs):
        """
        Load data from Form 460 models into OCD models.
        """
        self.header('Loading data extracted from Form 460 filings')

        # Only load committee types if we haven't already
        if not models.OCDCommitteeTypeProxy.objects.exists():
            self.log(' Creating committee types')
            self.load_model(models.OCDCommitteeTypeProxy)

        #
        # Committees
        #

        self.log(' Updating committee roster')
        self.load_model(models.OCDCommitteeProxy)
        self.load_model(models.OCDCommitteeIdentifierProxy)
        self.load_model(models.OCDCommitteeNameProxy)

        #
        # Filings
        #

        # self.load_model(models.OCDFilingProxy)
        # self.load_model(models.OCDFilingIdentifierProxy)

        #
        # Filing actions
        #

        # self.load_model(models.OCDFilingActionProxy)
        # self.load_model(models.OCDFilingActionSummaryAmountProxy)

        #
        # Transactions
        #

        # self.load_model(models.OCDTransactionProxy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Proxy models for OCD Filing related managers.
"""
from __future__ import unicode_literals
import logging
from psycopg2 import sql
from .base import CampaignFinanceManager
logger = logging.getLogger(__name__)


class OCDFilingManager(CampaignFinanceManager):
    """
    Manager with custom methods for OCD Filing model.
    """
    def _update_filings_from_form460_field(self, target_column, source_column):
        """
        Update target_column on OCD Filing with value from source_column.
        """
        from calaccess_processed_filings.models import Form460Filing, Filing

        source_table_identifier = sql.Identifier(Form460Filing._meta.db_table)
        source_column_identifier = sql.Identifier(Form460Filing._meta.get_field(source_column).column)
        target_column_identifier = sql.Identifier(Filing._meta.get_field(target_column).column)
        self.execute_custom_sql(
            'opencivicdata/campaign_finance/filings/update_filings_from_formfiling_table',
            composables={
                'source_table': source_table_identifier,
                'source_column': source_column_identifier,
                'target_column': target_column_identifier,
            },
        )

    def _update_filings_filer_id(self, model):
        """
        Update filer_id on OCD Filings.
        """
        processed_data_table_identifier = sql.Identifier(model._meta.db_table)
        self.execute_custom_sql(
            'opencivicdata/campaign_finance/filings/update_filings_filer_id',
            composables={
                'processed_data_table': processed_data_table_identifier,
            },
        )

    def load(self):
        """
        Load OCD Filing with data extracted from Form460Filing.
        """
        from calaccess_processed_filings.models import Form460Filing

        # Updating existing Filings...
        self._update_filings_from_form460_field('coverage_start_date', 'from_date')
        self._update_filings_from_form460_field('coverage_end_date', 'thru_date')
        self._update_filings_filer_id(Form460Filing)

        # Inserting new Filings...
        self.execute_custom_sql('opencivicdata/campaign_finance/filings/insert_filings_from_form460s')


class OCDFilingIdentifierManager(CampaignFinanceManager):
    """
    Manager with custom methods for OCD FilingIdentifier model.
    """
    def load(self):
        """
        Load OCD FilingIdentifier with data extracted from Form460Filing.
        """
        # Inserting new Filing Identifiers...
        self.execute_custom_sql('opencivicdata/campaign_finance/filings/insert_filing_ids_from_form460s')
        # Removing "calaccess_filing_id" from extras field on Filing...
        self.execute_custom_sql('opencivicdata/campaign_finance/filings/remove_calaccess_filing_ids_from_filings')


class OCDFilingActionManager(CampaignFinanceManager):
    """
    Manager with custom methods for OCD FilingAction model.
    """
    def load(self):
        """
        Load OCD FilingAction with data extracted from Form460FilingVersion.
        """
        # Inserting new Filing Actions...
        self.execute_custom_sql('opencivicdata/campaign_finance/filings/insert_filing_actions_from_form460s')

        # Setting is_current false on actions of amended Filings...
        self.execute_custom_sql('opencivicdata/campaign_finance/filings/set_is_current_for_old_filing_actions')

        # Setting is_current true on latest Filing Actions...
        self.execute_custom_sql('opencivicdata/campaign_finance/filings/set_is_current_for_new_filing_actions')


class OCDFilingActionSummaryAmountManager(CampaignFinanceManager):
    """
    Manager with custom methods for OCD FilingActionSummaryAmount model.
    """
    def _load_form460_summary_sheet_data(self):
        """
        Load OCD FilingActionSummaryAmount with data extracted from Form460FilingVersion.
        """
        from calaccess_processed_filings.models import Form460FilingVersion

        sum_fields = [
            f for f in Form460FilingVersion._meta.get_fields()
            if f.__class__.__name__ == 'IntegerField' and f.name[-3:] != '_id'
        ]
        table_composable = sql.Identifier(Form460FilingVersion._meta.db_table)
        for f in sum_fields:
            label_words = [w.capitalize() for w in f.name.split('_')]
            label = ' '.join(label_words)
            self.execute_custom_sql(
                'opencivicdata/campaign_finance/filings/insert_filing_action_summary_amounts',
                params={'label': label},
                composables={
                    'source_table': table_composable,
                    'source_column': sql.Identifier(f.column),
                }
            )

    def _load_from_form460_schedule_summary_field(self, model, label, field_name):
        """
        Load OCD FilingActionSummaryAmount with from a schedule summary field.
        """
        source_table_identifier = sql.Identifier(model._meta.db_table)
        source_column_identifier = sql.Identifier(model._meta.get_field(field_name).column)
        self.execute_custom_sql(
            'opencivicdata/campaign_finance/filings/insert_filing_action_summary_amount_from_schedules',
            params={'label': label},
            composables={
                'source_table': source_table_identifier,
                'source_column': source_column_identifier,
            },
        )

    def load(self):
        """
        Load OCD FilingActionSummaryAmount with data from all Form460 related models.
        """
        from calaccess_processed_filings.models import Form460ScheduleASummaryVersion, Form460ScheduleCSummaryVersion

        # Inserting new Filing Action Summary Amounts...
        # ...from Form 460 Summary Sheet...
        self._load_form460_summary_sheet_data()

        # ...from Form 460 Schedule A Summary...
        self._load_from_form460_schedule_summary_field(
            Form460ScheduleASummaryVersion,
            'Itemized Monetary Contributions',
            'itemized_contributions',
        )
        self._load_from_form460_schedule_summary_field(
            Form460ScheduleASummaryVersion,
            'Unitemized Monetary Contributions',
            'unitemized_contributions',
        )

        # ...from Form 460 Schedule C Summary...
        self._load_from_form460_schedule_summary_field(
            Form460ScheduleCSummaryVersion,
            'Itemized Non-monetary Contributions',
            'itemized_contributions',
        )
        self._load_from_form460_schedule_summary_field(
            Form460ScheduleCSummaryVersion,
            'Unitemized Non-monetary Contributions',
            'unitemized_contributions',
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Models for storing data from Campaign Disclosure Statements (Form 460).
"""
from django.db import models
from .base import Form460FilingBase


class Form460Filing(Form460FilingBase):
    """
    The most recent version of each Form 460 filing by recipient committees.

    Form 460 is the Campaign Disclosure Statement filed by all recipient
    committees, including:
    * Candidates, officeholders and their controlled committees
    * Primarily formed ballot measure committees
    * Primarily formed candidate/of ceholder committees
    * General purpose committees

    Recipient committes can use Form 460 to file:
    * Pre-election statements
    * Semi-annual statements
    * Quarterly statements
    * Termination statements
    * Special odd-year report

    Includes information from the cover sheet and summary page of the most
    recent version of each Form 460 filing. All versions of the filings can be
    found in Form460FilingVersion.
    """
    filing_id = models.IntegerField(
        verbose_name='filing id',
        primary_key=True,
        null=False,
        help_text='Unique identification number for the Form 460 filing ('
                  'from CVR_CAMPAIGN_DISCLOSURE_CD.FILING_ID)',
    )
    amendment_count = models.IntegerField(
        verbose_name='Count amendments',
        db_index=True,
        null=False,
        help_text='Number of amendments to the Form 460 filing (from '
                  'maximum value of CVR_CAMPAIGN_DISCLOSURE_CD.AMEND_ID)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        index_together = ((
            'filing_id',
            'amendment_count',
        ),)
        verbose_name = "Form 460 (Campaign Disclosure) filing"

    def __str__(self):
        return str(self.filing_id)


class Form460FilingVersion(Form460FilingBase):
    """
    Every version of each Form 460 (Campaign Disclosure Statement) filing by recipient committees.

    Includes information found on the cover sheet and summary page of each
    version of each Form 460 filing. For the most recent version of each filing,
    see Form460Filing.
    """
    filing = models.ForeignKey(
        'Form460Filing',
        related_name='versions',
        db_constraint=False,
        null=True,
        on_delete=models.SET_NULL,
        help_text='Unique identification number for the Form 460 filing ('
                  'from CVR_CAMPAIGN_DISCLOSURE_CD.FILING_ID)',
    )
    amend_id = models.IntegerField(
        verbose_name='amendment id',
        null=False,
        help_text='Identifies the version of the Form 460 filing, with 0 '
                  'representing the initial filing (from CVR_CAMPAIGN_'
                  'DISCLOSURE_CD.AMEND_ID)',
    )

    class Meta:
        """
        Model options.
        """
        app_label = 'calaccess_processed_filings'
        unique_together = ((
            'filing',
            'amend_id',
        ),)
        index_together = ((
            'filing',
            'amend_id',
        ),)
        verbose_name = "Form 460 (Campaign Disclosure) filing version"

    def __str__(self):
        return '%s-%s' % (self.filing, self.amend_id)



import os
import time
import calculate
from github import Github
from django.apps import apps
from django.conf import settings
from calaccess_processed.management.commands import CalAccessCommand
from django.contrib.humanize.templatetags.humanize import intcomma


class Command(CalAccessCommand):
    help = 'Analyze how many model fields lack documentation'

    def handle(self, *args, **kwargs):
        """
        Make it happen.
        """
        # Loop through all the models and find any fields without docs
        field_count = 0
        missing_list = []
        for m in self.get_model_list():
            field_list = m()._meta.fields
            field_count += len(field_list)
            for f in field_list:
                if not self.has_docs(f):
                    self.log("- [ ] {}".format(f))
                    missing_list.append((m, f))

        # If everything is done, declare victory
        if not missing_list:
            self.success("All %s fields documented!" % field_count)
            return False

        # If not, loop through the missing and create issues
        missing_count = len(missing_list)
        self.failure(
            "%s/%s (%d%%) of fields lack documentation" % (
                intcomma(missing_count),
                intcomma(field_count),
                calculate.percentage(missing_count, field_count)
            )
        )

    def get_model_list(self):
        return apps.get_app_config('elections').models.values() + apps.get_app_config('core').models.values()

    def has_docs(self, field):
        """
        Test if a Django field has some kind of documentation already.

        Returns True or False
        """
        if field.name == 'id':
            return True
        if field.help_text:
            return True
        if field.__dict__['_verbose_name']:
            return True
        return False
